/*
 * GisUtils.java
 *
 * Copyright (c) 2005 Bernhard Fischer.
 * This file is part of TopGMap. TopGMap is free software; it is
 * published under the terms of the GNU General Public License.
 */

package de.bfisch.gis;

import java.awt.geom.Point2D;


public class GisUtils {
    
    
    private static class Ellipse {
        public String name;
        public double a;
        public double invf;
        
        public Ellipse(String name, double a, double invf) {
            this.name = name;
            this.a    = a;
            this.invf = invf;
        }
    }
    
    
    private static class CrsDist {
        public double crs12;
        public double crs21;
        public double d;
    }
    
    
    public static double dist(Point2D pos1, Point2D pos2) {
        if (pos1.equals(pos2)) {
            return 0;
        }
        
        double glat1 =   Math.PI / 180d * pos1.getX();
        double glon1 = - Math.PI / 180d * pos1.getY();
        double glat2 =   Math.PI / 180d * pos2.getX();
        double glon2 = - Math.PI / 180d * pos2.getY();
        Ellipse ellipse = new Ellipse("WGS84", 6378.137 / 1.852, 298.257223563);
        
        CrsDist crsDist = crsDist(glat1, glon1, glat2, glon2, ellipse);
        
        return crsDist.d * 1000;
    }
            
    
    /**
     * http://williams.best.vwh.net/gccalc.htm
     */
    public static CrsDist crsDist(double glat1, double glon1, double glat2, double glon2, Ellipse ellipse) {
        // glat1 initial geodetic latitude in radians N positive 
        // glon1 initial geodetic longitude in radians E positive 
        // glat2 final geodetic latitude in radians N positive 
        // glon2 final geodetic longitude in radians E positive 
        double a = ellipse.a;
        double f = 1 / ellipse.invf;
        //Logger.debug("a=" + a + ", f=" + f);
        //var r, tu1, tu2, cu1, su1, cu2, s1, b1, f1
        //var x, sx, cx, sy, cy,y, sa, c2a, cz, e, c, d
        double EPS= 0.00000000005;
        double faz, baz, s;
        int iter = 1;
        int MAXITER = 100;
        if (glat1 + glat2 == 0. && Math.abs(glon1 - glon2) == Math.PI){
            System.err.println("Course and distance between antipodal points is undefined");
            glat1 = glat1 + 0.00001; // allow algorithm to complete
        }
        if (glat1 == glat2 && (glon1 == glon2 || Math.abs(Math.abs(glon1 - glon2) - 2 * Math.PI) < EPS)){
            System.err.println("Points 1 and 2 are identical- course undefined");
            CrsDist out = new CrsDist();
            out.d = 0;
            out.crs12 = 0;
            out.crs21 = Math.PI;
            return out;
        }
        double r = 1 - f;
        double tu1 = r * Math.tan(glat1);
        double tu2 = r * Math.tan(glat2);
        double cu1 = 1. / Math.sqrt(1. + tu1 * tu1);
        double su1 = cu1 * tu1;
        double cu2 = 1. / Math.sqrt(1. + tu2 * tu2);
        double s1 = cu1 * cu2;
        double b1 = s1 * tu2;
        double f1 = b1 * tu1;
        double x = glon2 - glon1;
        //Logger.debug("x=" + x);
        double y = 0.;
        double d = x + 1; // force one pass
        double sx = 0.;
        double sy = 0.;
        double cx = 0.;
        double cy = 0.;
        double cz = 0.;
        double c2a = 0.;
        double e = 0.;
        while (Math.abs(d - x) > EPS && iter < MAXITER) {
            iter = iter + 1;
            sx = Math.sin(x);
            //Logger.debug("sx=" + sx);
            cx = Math.cos(x);
            tu1 = cu2 * sx;
            tu2 = b1 - su1 * cu2 * cx;
            sy = Math.sqrt(tu1 * tu1 + tu2 * tu2);
            cy = s1 * cx + f1;
            y = Math.atan2(sy, cy);
            double sa = s1 * sx / sy;
            c2a = 1 - sa * sa;
            cz = f1 + f1;
            if (c2a > 0.)
                cz = cy - cz / c2a;
            e = cz * cz * 2. - 1.;
            double c = ((-3. * c2a + 4.) * f + 4.) * c2a * f / 16.;
            d = x;
            x = ((e * cy * c + cz) * sy * c + y) * sa;
            x = (1. - c) * x * f + glon2 - glon1;
        }
        //Logger.debug("tu1=" + tu1 + ", tu2=" + tu2);
        faz = modcrs(Math.atan2(tu1, tu2));
        baz = modcrs(Math.atan2(cu1 * sx, b1 * cx - su1 * cu2) + Math.PI);
        x = Math.sqrt ((1 / (r * r) - 1) * c2a + 1);
        x += 1;
        x = (x - 2.) / x;
        double c = 1. - x;
        c = (x * x / 4. + 1.) / c;
        d = (0.375 * x * x - 1.) * x;
        x = e * cy;
        s = ((((sy * sy * 4. - 3.) * (1. - e - e) * cz * d / 6. - x) * d / 4. + cz) * sy * d + y) * c * a * r;
        CrsDist out = new CrsDist();
        out.d = s;
        out.crs12 = faz;
        out.crs21 = baz;
        if (Math.abs(iter-MAXITER) < EPS) {
            System.err.println("Algorithm did not converge");
        }
        return out;
    }
    

    private static double modcrs(double x) {
        x = x % (2 * Math.PI);
        if (x < 0) {
            x += 2 * Math.PI;
        }
        return x;
    }
    
    
    /*
    public static void main(String[] args) {
        Ellipse ellipse = new Ellipse("WGS84", 6378.137 / 1.852, 298.257223563);
        
        // Mnchen: Lat/Lon: 48.1 N 11.7 E
        // Vorzeichen: N/W: +, S/E: -
        double lat0 =  47.47250174;
        double lon0 = -10.90005897;

        double lat1 =  47.46910039;
        double lon1 = -10.88779354;

        double lat2 =  47.47769730;
        double lon2 = -10.88389732;

        double lat3 =  47.47512210;
        double lon3 = -10.89662133;
        
        double lat4 =  47.6039993;
        double lon4 = -11.88421565;
        
        double lat5 =  47.60795327;
        double lon5 = -11.88081547;
        
        lat1 = Math.PI / 180d * lat4;
        lon1 = Math.PI / 180d * lon4;

        lat2 = Math.PI / 180d * lat5;
        lon2 = Math.PI / 180d * lon5;
        
        System.out.println("lat1=" + lat1 + ", lon1=" + lon1 + ", lat2=" + lat2 + ", lon2="+ lon2);

        CrsDist crsDist = crsDist(lat1, -lon1, lat2, -lon2, ellipse);
        
        double crs12 = crsDist.crs12 * 180d / Math.PI;
        double crs21 = crsDist.crs21 * 180d / Math.PI;
        double dc = 1.852;    // km
        double d = crsDist.d * dc;
        
        System.out.println(  "crs12=" + crs12 +
                           ", crs21=" + crs21 +
                           ", d=" + d);
    }
    */
    
}