/*
 * GpsIdGenerator.java
 *
 * Copyright (c) 2005 Bernhard Fischer.
 * This file is part of TopGMap. TopGMap is free software; it is
 * published under the terms of the GNU General Public License.
 */

package de.bfisch.gis;

import java.util.Iterator;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;


/**
 * Id generator.
 */
public class GpsIdGenerator {
        
    
    /**
     * Generator for uniform, unique ids. 
     */
    class IdGenerator {
        private static final int TYPE_NUMERIC = 1;
        private static final int TYPE_ALPHA   = 2;
        
        /** Successor id generator, work is delegated to this instance */
        private IdGenerator succGenerator;
        
        private String baseId;
        private int idLen;
        private String separator;
        private int seqNo;
        private int seqNoLen;
        private int seqNoType;
	
        /**
         * Create an new id generator.
         */
	public IdGenerator(String baseId, int idLen, String separator, int seqNo, int seqNoLen, int seqNoType) {
            this.baseId    = baseId;
            this.idLen     = idLen;
            this.separator = separator;
            this.seqNo     = seqNo;
            this.seqNoLen  = seqNoLen;
            this.seqNoType = seqNoType;
	}
        
        /**
         * Return the next more general id generator for the given id.
         */
        public IdGenerator nextIdGenerator(String id) {
            IdGenerator idGenerator = getIdGenerator(id, idLen, seqNoLen + 1, seqNo);
            setSuccIdGenerator(idGenerator);
            idGenerator.addSeqNo(seqNo);
            return idGenerator;
        }
        
        /**
         * Set the successor id generator.
         */
        public void setSuccIdGenerator(IdGenerator idGenerator) {
            succGenerator = idGenerator;
        }
        
        /**
         * Increase current sequence number.
         */
        public void addSeqNo(int seqNo) {
            this.seqNo += seqNo;
        }
	
        /**
         * Return next unique id (or null if failed).
         */
	public String nextId() {
            // Initialize id
	    StringBuffer uniqueId = new StringBuffer(idLen);
	    //uniqueId.append(id);
            uniqueId.append(baseId);
            
            // Truncate id to base length
	    int baseLen = idLen - separator.length() - seqNoLen;
	    if (uniqueId.length() > baseLen)
		uniqueId.setLength(baseLen);
            
            // Append separator
	    uniqueId.append(separator);
            
            // Append sequence number
	    switch (seqNoType) {
	    case TYPE_NUMERIC:
		appendSeqNo(uniqueId, seqNo, '0', 10, seqNoLen);
		break;
	    case TYPE_ALPHA:
		appendSeqNo(uniqueId, seqNo, 'A', 26, seqNoLen);
		break;
	    }
            seqNo++;
            
            // Return result (null if failed)
	    if (uniqueId.length() > 0) {
		return uniqueId.toString();
            } else {
                seqNo--;
		return null;
            }
	}
	
        /**
         * Append sequence number to id.
         */
	private void appendSeqNo(StringBuffer uniqueId, int seqNo, char baseChar, int baseSeqNo, int seqNoLen) {
	    uniqueId.setLength(uniqueId.length() + seqNoLen);
	    for (int i = 0; i < seqNoLen; i++) {
		int divSeqNo = seqNo / baseSeqNo;
		int modSeqNo = seqNo % baseSeqNo;
		char c = (char)(baseChar + modSeqNo);
		uniqueId.setCharAt(uniqueId.length() - 1 - i, c);
		seqNo = divSeqNo;
	    }
	    if (seqNo != 0) {
		// seqNo doesn't fit to seqNoLen digits
		uniqueId.setLength(0);
	    }
	}
        
        /**
         * Reset current sequence number.
         */
        public void reset() {
            seqNo = 0;
        }
	
    } // IdGenerator
    
    
    /** Mapping of base id to id generator */
    private Map<String, IdGenerator> idGeneratorMap = new HashMap<String, IdGenerator>();
    
    /** Pool with all ids (superfluous) */
    private Set<String> idPool = new HashSet<String>();

    private int idLen;
    private String defaultId;
    
    
    /**
     * Create an id generator.
     */
    public GpsIdGenerator() {
        this(null, Integer.MAX_VALUE, 1);
    }
    
    
    /**
     * Create an id generator.
     */
    public GpsIdGenerator(String defaultId, int idLen, int digits) {
        this.idLen = idLen;
        setDefaultId(defaultId);
    }
    
    
    /**
     * Set the default id.
     */
    private void setDefaultId(String defaultId) {
        this.defaultId = defaultId;
        if (defaultId != null) {
            norm(defaultId);
        }
    }
    
    
    /**
     * Remove all generated ids.
     * Id idGeneratorors will be preserved.
     */
    public void reset() {
        for (Iterator iter = idGeneratorMap.values().iterator(); iter.hasNext(); ) {
            IdGenerator idGenerator = (IdGenerator)iter.next();
            idGenerator.reset();
        }
        idPool.clear();
    }
    
    
    /**
     * Reset complete id generator.
     */
    public void clear() {
        idGeneratorMap.clear();
        idPool.clear();
    }

    
    /**
     * 
     */
    public String defaultId() {
        return norm(defaultId);
    }
    
    
    /**
     * Normalize id;
     * it is assured that generated id is unique. 
     */
    public String norm(String id) {
        // Normalize id
        id = reviseId(id);
        
        // Find id generator
        IdGenerator idGenerator = getIdGenerator(id, idLen);

        String uniqueId = null;
        
        do {
            // Generate unique id
            uniqueId = idGenerator.nextId();
            
            // Generation failed?
            if (uniqueId == null) {
        	// Try next id generator
        	idGenerator = idGenerator.nextIdGenerator(id);
            }

            // Generated id should be unique
            //assert uniqueId == null || !idPool.contains(uniqueId);
            boolean passed = uniqueId == null || !idPool.contains(uniqueId);
            if (!passed)
                System.out.println("failed");
        } while (uniqueId == null || uniqueId.length() == 0 || idPool.contains(uniqueId));

        // Add generated id to id pool
        idPool.add(uniqueId);
        
        return uniqueId;
    }
    
    
    /**
     * Replace/remove invalid characters.
     */
    private String reviseId(String id) {
        // Initialize id
        StringBuffer baseIdBuf = new StringBuffer(id != null ? id : defaultId != null ? defaultId : "");
        
        // Replace/remove invalid characters
        for (int i = baseIdBuf.length() - 1; i >= 0 ; i--) {
            char c = baseIdBuf.charAt(i);
            if ((c >= 'A' && c <= 'Z') ||
                (c >= '0' && c <= '9') ||
                 c == ' ' || c == '_'  ||
                 c == '+' || c == '-'     ) {
                // OK, nothing to do
            } else
            if ( c >= 'a' && c <= 'z') {
                baseIdBuf.setCharAt(i, Character.toUpperCase(c));
            } else
            if ( c == '' || c == '') {
                baseIdBuf.setCharAt(i, 'A');
            } else
            if ( c == '' || c == '') {
                baseIdBuf.setCharAt(i, 'O');
            } else
            if ( c == '' || c == '') {
                baseIdBuf.setCharAt(i, 'U');
            } else
            if ( c == '') {
                baseIdBuf.setCharAt(i, 'S');
            } else {
                baseIdBuf.deleteCharAt(i);
            }
        }
        
        // Truncate id
        if (baseIdBuf.length() > idLen) {
            baseIdBuf.setLength(idLen);
        }
        
        return baseIdBuf.toString();
    }
    
    
    /**
     * Return an id generator for the given id.
     */
    public IdGenerator getIdGenerator(String id, int idLen) {
        IdGenerator idGenerator = getIdGenerator(id, idLen, 0, 0);
        return idGenerator;
    }
    
    
    /**
     * Return an id generator for the given id.
     */
    private IdGenerator getIdGenerator(String id, int idLen, int seqNoLen, int seqNo) {
        // Fix base id
        int baseIdLen = idLen - seqNoLen;
        if (baseIdLen > 2 && seqNoLen > 0)
            baseIdLen -= 1;
        String baseId = id;
        if (baseId.length() > baseIdLen)
            baseId = baseId.substring(0, baseIdLen);

        // Reuse id generator for base id if available
        IdGenerator idGenerator = getMappedIdGenerator(baseId, seqNoLen);
        
        if (idGenerator != null) {
            // Return found id generator
            return idGenerator;
        }
        
        // Create new id generator for base id
        char lastChar = baseId.length() > 0 ? baseId.charAt(baseId.length() - 1) : 'A';
        int seqNoType = Character.isLetter(lastChar) ? IdGenerator.TYPE_NUMERIC : IdGenerator.TYPE_ALPHA;
        String separator = baseId.length() > 1 && seqNoLen > 0 ? "-" : "";
        idGenerator = new IdGenerator(baseId, idLen, separator, 0, seqNoLen, seqNoType);
        putMappedIdGenerator(baseId, idGenerator);
    
        return idGenerator;
    }
    
    
    /**
     * Retrieve id generator from the map.
     */
    private void putMappedIdGenerator(String baseId, IdGenerator idGenerator) {
        idGeneratorMap.put(baseId, idGenerator);
    }
    
    
    /**
     * Retrieve id generator from the map.
     */
    private IdGenerator getMappedIdGenerator(String baseId, int seqNoLen) {
        IdGenerator idGenerator = (IdGenerator)idGeneratorMap.get(baseId);
        if (idGenerator != null && idGenerator.seqNoLen >= seqNoLen) {
            // There is a valid id generator with sufficient sequence number length.
            // Use successor generator if available.
            while (idGenerator.succGenerator != null)
                idGenerator = idGenerator.succGenerator;
            return idGenerator;
        }
        return null;
    }
    
    
    /**
     * Remove an id.
     */
    public boolean remove(String id) {
        // @todo decrement sequence number of id generator, possibly revert to more special id generator.
        return idPool.remove(id);
    }
    
    
    public static void main(String[] args) {
        GpsIdGenerator testPool = new GpsIdGenerator(null, 6, 2);
        
        System.out.println("PHASE 1");
        
        //String id = "DH";
        //String id = "D$7Htte";
        //String id = "DS7HUT";
        String id = "DS7H";
        for (int i = 0; i < 25; i++) {
            //System.out.println("[" + i + "] " + testPool.defaultId());
            System.out.println("[" + i + "] " + testPool.norm(id));
        }

        String id2 = "DS700";
        for (int i = 0; i < 32; i++) {
            System.out.println("[" + i + "] " + testPool.norm(id2));
        }
        
        for (int i = 0; i < 7; i++) {
            System.out.println("[" + i + "] " + testPool.norm(id));
        }
        
        System.out.println();
        System.out.println("PHASE 2");
        
        testPool.reset();

        for (int i = 0; i < 25; i++) {
            System.out.println("[" + i + "] " + testPool.norm(id));
        }
        
        for (int i = 0; i < 32; i++) {
            System.out.println("[" + i + "] " + testPool.norm(id2));
        }
        
        for (int i = 0; i < 7; i++) {
            System.out.println("[" + i + "] " + testPool.norm(id));
        }
    }
    
}