/*
 * GpsWaypoint.java
 *
 * Copyright (c) 2005 Bernhard Fischer.
 * This file is part of TopGMap. TopGMap is free software; it is
 * published under the terms of the GNU General Public License.
 */

package de.bfisch.gis;

import java.awt.geom.Point2D;

import de.bfisch.geom.Point2DSer;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Set;
import java.util.HashSet;

import de.bfisch.util.Logger;
import de.bfisch.util.Propertable;


public class GpsWaypoint extends GpsElement implements Propertable {
    
    public static final int TYPE_WAYPOINT       = 11;
    public static final int TYPE_ROUTE_LABEL    = 12;
    public static final int TYPE_TRACK_LABEL    = 13;
    public static final int TYPE_DELETED        = 14;

    public static final int WAYPOINT_NAMED      = 21;
    public static final int WAYPOINT_REFERENCE  = 22;
    public static final int WAYPOINT_UNNAMED    = 23;
    
    public static final int ROUTE_BEGIN         = 31;
    public static final int ROUTE_END           = 32;
    
    private static final Logger log = new Logger("GpsWaypoint");
    
    //private static GpsIdPool waypointNamePool = new GpsIdPool(null, 6, 2);
    private static GpsIdGenerator waypointNamePool = new GpsIdGenerator(null, 6, 2);
    private static List<GpsWaypoint> gpsWaypoints = new ArrayList<GpsWaypoint>();
    
    private static int nextId;
    
    private int id;
    private Point2D position;
    private String name;
    private String gpsName;
    
    private int type;
    private int typeWaypoint;
    private int typeRoute;
    
    private transient Set<GpsWaypoint> neighbours;
    
    
    private GpsWaypoint(Point2D position, int typeWaypoint, String source) {
        this(position, typeWaypoint, null, source);
    }
    
    
    private GpsWaypoint(Point2D position, int typeWaypoint, String name, String source) {
        super(source);
        this.id           = nextId++;
        this.position     = new Point2DSer(position);
        this.type         = TYPE_WAYPOINT;
        this.typeWaypoint = typeWaypoint;
        if (typeWaypoint == WAYPOINT_NAMED) {
            setName(name);
        }
        log.debug("Waypoint created.",
                new Object[] {
                   "Waypoint 1", this,
        });
    }
    
    
    public static GpsWaypoint newInstance(Point2D position) {
        return newInstance(position);
    }
    
    
    public static GpsWaypoint newInstance(Point2D position, String source) {
        GpsWaypoint gpsWaypoint = new GpsWaypoint(position, WAYPOINT_UNNAMED, source);
        gpsWaypoints.add(gpsWaypoint);
        return gpsWaypoint;
    }
    
    
    public static GpsWaypoint newReferenceInstance(Point2D position, String source) {
        GpsWaypoint gpsWaypoint = new GpsWaypoint(position, WAYPOINT_REFERENCE, source);
        gpsWaypoints.add(gpsWaypoint);
        return gpsWaypoint;
    }
    
    
    public static GpsWaypoint newInstance(Point2D position, String name, String source) {
        GpsWaypoint gpsWaypoint = new GpsWaypoint(position, WAYPOINT_NAMED, name, source);
        gpsWaypoints.add(gpsWaypoint);
        return gpsWaypoint;
    }
    
    
    public static List instances() {
        return gpsWaypoints;
    }

    
    public static void reset() {
        waypointNamePool.reset();
        for (Iterator iterWaypoint = gpsWaypoints.iterator(); iterWaypoint.hasNext(); ) {
            GpsWaypoint gpsWaypoint = (GpsWaypoint)iterWaypoint.next();
            gpsWaypoint.resetGpsName();
        }
    }
    
    
    public void resetGpsName() {
        gpsName = null;
    }
    
    
    public static void clear() {
        waypointNamePool.clear();
        gpsWaypoints.clear();
    }
    
    
    public boolean delete() {
        assert type != TYPE_DELETED;
        type = TYPE_DELETED;
        //boolean success = gpsWaypoints.remove(this);
        if (neighbours != null) {
            for (Iterator iterNeighbours = neighbours.iterator(); iterNeighbours.hasNext(); ) {
                GpsWaypoint neighbour = (GpsWaypoint)iterNeighbours.next();
                neighbour.removeNeighbour(this);
            }
            neighbours = null;
        }
        log.debug("Waypoint deleted.",
                new Object[] {
                   "Waypoint 1", this,
        });
        return true;
    }
    
    
    public static int nextId() {
        return nextId;
    }
    
    
    public int getId() {
        assert !isDeleted();
        return id;
    }
    
    
    public void setPosition(Point2D position) {
        assert !isDeleted();
        this.position = new Point2DSer(position);
    }
    
    
    public Point2D getPosition() {
        assert !isDeleted();
        return position;
    }
    
    
    public void setName(String name) {
        assert !isDeleted();
        if (name != null) {
            char firstChar = name.charAt(0);
            switch (firstChar) {
            case '*':
                this.type         = TYPE_WAYPOINT;
                this.typeWaypoint = WAYPOINT_NAMED;
                this.typeRoute    = ROUTE_BEGIN;
                name = name.substring(1);
                break;
            case '+':
                this.type         = TYPE_WAYPOINT;
                this.typeWaypoint = WAYPOINT_NAMED;
                this.typeRoute    = ROUTE_END;
                name = name.substring(1);
                break;
            case '/':
                this.type         = TYPE_ROUTE_LABEL;
                name = name.substring(1);
                break;
            case ':':
                this.type         = TYPE_TRACK_LABEL;
                name = name.substring(1);
                break;
            default:
                this.type         = TYPE_WAYPOINT;
                this.typeWaypoint = WAYPOINT_NAMED;
                break;
            }
        } else {
            this.type         = TYPE_WAYPOINT;
            // typeWaypoint: WAYPOINT_REFERENCE || WAYPOINT_UNNAMED
        }
        this.name = name;
    }
    
    
    public String getName() {
        assert !isDeleted();
        return name;
    }
    
    
    public String getGpsName() {
        assert !isDeleted();
        if (gpsName == null) {
            if (name == null) {
                // Prefix for unnamed waypoints
                switch (neighbours.size()) {
                case 0:
                    break;
                case 1:
                    name = "T";
                    break;
                case 2:
                    name = "I";
                    break;
                case 3:
                    name = "Y";
                    break;
                default:
                    name = "X";
                    break;
                }
            }
            gpsName = waypointNamePool.norm(name);
        }
        return gpsName;
    }
    
    
    public int getAccuracy() {
        assert !isDeleted();
        switch (typeWaypoint) {
            case WAYPOINT_REFERENCE:
                return 3;
            case WAYPOINT_NAMED:
                return 2;
            case WAYPOINT_UNNAMED:
                return 1;
        }
        return -1;
    }
    
    
    public boolean isWaypoint() {
        return type == TYPE_WAYPOINT;
    }
    
    
    public boolean isRouteLabel() {
        return type == TYPE_ROUTE_LABEL;
    }
    
    
    public boolean isTrackLabel() {
        return type == TYPE_TRACK_LABEL;
    }
    
    
    public boolean isDeleted() {
        return type == TYPE_DELETED;
    }
    
    
    public void setNamed() {
        assert isWaypoint();
        if (name == null) {
            setName(null);
        }
        typeWaypoint = WAYPOINT_NAMED;
    }
    
    
    public boolean isNamed() {
        return isWaypoint() &&
               typeWaypoint == WAYPOINT_NAMED;
    }
    
    
    public boolean isReference() {
        return isWaypoint() &&
               typeWaypoint == WAYPOINT_REFERENCE;
    }
    
    
    public boolean isUnnamed() {
        return isWaypoint() &&
               typeWaypoint == WAYPOINT_UNNAMED;
    }
    
    
    public boolean isBegin() {
        return isNamed() &&
               typeRoute == ROUTE_BEGIN;
    }
    
    
    public boolean isEnd() {
        return isNamed() &&
               typeRoute == ROUTE_END;
    }
    
    
    public void addNeighbour(GpsWaypoint gpsWaypoint) {
        assert !isDeleted();
        if (neighbours == null) {
            neighbours = new HashSet<GpsWaypoint>();
        }
        neighbours.add(gpsWaypoint);
    }
    
    
    public void removeNeighbour(GpsWaypoint gpsWaypoint) {
        assert !isDeleted();
        if (neighbours == null) {
            return;
        }
        neighbours.remove(gpsWaypoint);
    }
    
    
    public Collection neighbours() {
        assert !isDeleted();
        return neighbours;
    }
    
    
    public Object[] properties() {
        int fixLength = 6 * 2;
        int entries = fixLength;
        Object[] props = new Object[entries];
        int i = 0;
        props[i++] = "id";
        props[i++] = new Integer(id);
        props[i++] = "type";
        if (isRouteLabel() || isTrackLabel()) {
            props[i++] = "label";
        } else
        if (isReference()) {
            props[i++] = "ref";
        } else
        if (isDeleted()) {
            props[i++] = "deleted";
        } else
        if (isNamed()) {
            props[i++] = "named";
        } else {
            props[i++] = "unnamed";
        }
        props[i++] = "name";
        props[i++] = name;
        props[i++] = "lat";
        props[i++] = new Double(position.getX());
        props[i++] = "lon";
        props[i++] = new Double(position.getY());
        props[i++] = "source";
        props[i++] = getSource();
        return props;
    }

    
    public String toString() {
        StringBuffer buffer = new StringBuffer("GpsWaypoint@");
        buffer.append(Integer.toHexString(hashCode()));
        Object[] props = properties();
        if (props != null) {
            int n = 0;
            for (int i = 0; i < props.length; i += 2) {
                Object key   = props[i];
                Object value = props[i + 1];
                if (value != null) {
                    buffer.append(n == 0 ? "[" : ",");
                    n++;
                    buffer.append(key);
                    buffer.append("=");
                    buffer.append(value);
                }
            }
            if (n > 0)
                buffer.append("]");            
        }
        return buffer.toString();
    }

}
