/*
 * OvlReader.java
 *
 * Copyright (c) 2005 Bernhard Fischer.
 * This file is part of TopGMap. TopGMap is free software; it is
 * published under the terms of the GNU General Public License.
 */

package de.bfisch.gis;

import java.awt.geom.Point2D;
import java.io.IOException;
import java.io.Reader;
import java.util.prefs.Preferences;

import org.ini4j.IniPreferences;

import de.bfisch.util.Logger;


/**
 * Gebilde.
 */
public class OvlReader {
    
    private static final Logger log = new Logger("OvlReader");
    
    private Reader input;
    private boolean inited = false;
    private Preferences prefs;
    private int iSymbol;
    private int nSymbols;
    
    
    public OvlReader(Reader input) {
        this.input = input;
    }
    
    
    private void init() throws IOException {
        inited = true;
        prefs = new IniPreferences(input);
        Preferences deck = prefs.node("Overlay");
        nSymbols = (int)deck.getLong("Symbols", -1);
        log.debug("Total symbols=" + nSymbols);
        iSymbol = 0;
    }
    
    
    public GpsElement nextElement() throws IOException {
        if (!inited) {
            init();
        }
        
        GpsElement gpsElement = null;
        
        while (gpsElement == null && iSymbol < nSymbols) {
            iSymbol++;
            String deckName = "Symbol " + Integer.toString(iSymbol);
            Preferences deck = prefs.node(deckName);
            int type = deck.getInt("Typ", -1);
    
            double lat;
            double lon;
            Point2D position;
            
            switch (type) {
                case 1:        // Bitmap
                case 5:        // Rechteck
                case 6:        // Kreis
                case 7:        // Dreieck
                    lon = deck.getDouble("XKoord", 0.000);
                    lat = deck.getDouble("YKoord", 0.000);
                    //Logger.debug("[" + deckName + "] lat=" + lat + ", lon=" + lon);
                    position = new Point2D.Double(lat, lon);
                    gpsElement = GpsWaypoint.newReferenceInstance(position, deckName);
                    break;
                    
                case 2:        // Text
                    lon = deck.getDouble("XKoord", 0.000);
                    lat = deck.getDouble("YKoord", 0.000);
                    //Logger.debug("[" + deckName + "] lat=" + lat + ", lon=" + lon);
                    position = new Point2D.Double(lat, lon);
                    String text = deck.get("Text", null);
                    gpsElement = GpsWaypoint.newInstance(position, text, deckName);
                    break;
                    
                case 3:        // Linie
                case 4:        // Polygon
                    log.debug("Line follows.");
                    int nPositions = deck.getInt("Punkte", 0);
                    GpsLine gpsLine = GpsLine.newInstance(deckName);
                    for (int k = 0; k < nPositions; k++) {
                        String coordName = "Koord" + Integer.toString(k);
                        String source = deckName + "/" + coordName;
                        lon = deck.getDouble("X" + coordName, 0.000);
                        lat = deck.getDouble("Y" + coordName, 0.000);
                        //Logger.debug("  lat=" + lat + ", lon=" + lon);
                        position = new Point2D.Double(lat, lon);
                        GpsWaypoint gpsWaypoint = GpsWaypoint.newInstance(position, source);
                        gpsLine.add(gpsWaypoint);
                    }
                    log.debug("Line assembled.",
                            new Object[] {
                               "Line", gpsLine,
                    });
                    gpsElement = gpsLine;
                    break;
        
                default:    // unknown
                    System.err.println("WARNING: Overlay object " + deckName + " of unknown type " + type + " skipped.");
                    // continue with next overlay deck
                    break;
                
            }
        }
        
        return gpsElement;
    }
    
}
