/*
 * TopGMap.java
 *
 * Copyright (c) 2005 Bernhard Fischer.
 * This file is part of TopGMap. TopGMap is free software; it is
 * published under the terms of the GNU General Public License.
 */

package de.bfisch.gis;

import de.bfisch.io.FileUtils;
import de.bfisch.io.NullOutputStream;
import de.bfisch.math.Geometry;
import de.bfisch.util.Logger;

import java.awt.geom.Point2D;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;


/**
 * Vorgehen:
 * 1. Einlesen aller Strecken (= Linie), Texte und Kreise / Rechtecke
 * 2. Koordinaten-Korrekturlauf
 *    Ausrichtung: Kreis <- Rechteck <- Text <- Linie
 * 3a Routen sind mit /Name gekennzeichnet, Tracks mit :Name
 * 3b Anfangspunkte sind mit *Name gekennzeichnet, Endpunkte mit +Name
 * 4. Streckennamen den Strecken zuordnen (Abstand Punkt/Gerade < eps)
 * 5a Namen von Wegpunkten normalisieren:
 *    Groschreibung, keine Umlaute und Sonderzeichen, etc.
 * 5b Wegpunkte mit gleichem Namen durchnummerieren,
 *    z. B. JH-1, JH-2 bzw. 1600-A, 1600-B
 * 5c Wegpunkte der Routen generieren, soweit nicht schon vorhanden;
 *    Schema: <erster Buchstabe Routenname><Nummer> bzw.
 *    bei Kreuzung X<erster Bu Route1><erster Bu Route2>
 *    oder X<erster Bu Route1><Nummer>
 * 6. Die n lngsten Routen exportieren
 *    Route: Anfangs- und Endpunkt sind bezeichnet
 * 7. Die n lngsten Tracks exportieren
 *    Track: Alle Strecken, auch die Routen
 * 8. Basemap zusammenstellen:
 *    Track, welcher alle Strecken durchluft
 * 9. Basemap exportieren
 * 
 * Zu tun:
 * 1. Klasse Waypoint berarbeiten
 * 2. Algorithmus finden, der alle Strecken durchluft (Traveling Salesman Problem / Handelsreisender-Problem)
 * 
 * Kapazitt Geko 201:
 * -  Waypoints/icons: 500 with name and graphic symbol
 * -  Routes: 20 reversible x 125 legs/route
 * -  Tracks: 10,000 trackpoints and 10 saved tracks; automatic track log; TracBack lets you retrace your path
 */
public class TopGMap {
    
    public static final String VERSION = "1.0.0";
    
    private static final Logger log = new Logger("TopGMap");
    
    private TopGMapConfig config;
    
    public boolean autoAccuracy;                        // Automatische Bestimmung der Positionsgenauigkeit
    public double posAccuracy;                          // Positionsgenauigkeit in Meter
    public double posAccuracyWarn;                      // Warnung, wenn unterschritten wird
    public int trackMaxIncrement        = 5000;         // Maximale Schrittweite fr Tracks in Meter
    
    public boolean ignoreLimits;                        // Gerte-Einschrnkungen ignorieren
    public int maxWaypoints             = 500;          // Maximale Anzahl Waypoints
    public int maxRoutes                = 20;           // Maximale Anzahl Routen
    public int maxTracks                = 10;           // Maximale Anzahl Tracks
    
    
    public TopGMap() {
        init();
    }
    
    
    private void init() {
        config = TopGMapConfig.getInstance("TopGMap");
        
        if (config.getName().equals("TopGMap")) {
            try {
                config.load();
            } catch (IOException ex) {
                log.warn("Konfiguration kann nicht geladen werden",
                	 new Object[] { "Exception", ex });
            }
        }
        
        applyConfig();
    }
    
    
    public void close() {
        extractConfig();
        
        if (config.getName().equals("TopGMap")) {
            try {
                config.store();
            } catch (IOException ex) {
                log.warn("Konfiguration kann nicht gespeichert werden",
                	 new Object[] { "Exception", ex });
            }
        }
    }
    
    
    private void applyConfig() {
        config.addPropertyChangeListener("autoAccuracy", new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent e) {
                String value = (String)e.getNewValue();
                autoAccuracy = TopGMapConfig.parseAutoAccuracy(value);
            }
        }, true);
        config.addPropertyChangeListener("posAccuracy", new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent e) {
                String value = (String)e.getNewValue();
                posAccuracy = TopGMapConfig.parsePosAccuracy(value);
            }
        }, true);
        config.addPropertyChangeListener("posAccuracyWarn", new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent e) {
                String value = (String)e.getNewValue();
                posAccuracyWarn = TopGMapConfig.parsePosAccuracyWarn(value);
            }
        }, true);
        trackMaxIncrement = config.getTrackMaxIncrement();
        
        config.addPropertyChangeListener("ignoreLimits", new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent e) {
                String value = (String)e.getNewValue();
                ignoreLimits = TopGMapConfig.parseIgnoreLimits(value);
            }
        }, true);
        config.addPropertyChangeListener("maxWaypoints", new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent e) {
                String value = (String)e.getNewValue();
                maxWaypoints = TopGMapConfig.parseMaxWaypoints(value);
            }
        }, true);
        config.addPropertyChangeListener("maxRoutes", new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent e) {
                String value = (String)e.getNewValue();
                maxRoutes = TopGMapConfig.parseMaxRoutes(value);
            }
        }, true);
        config.addPropertyChangeListener("maxTracks", new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent e) {
                String value = (String)e.getNewValue();
                maxTracks = TopGMapConfig.parseMaxTracks(value);
            }
        }, true);
        
        config.addPropertyChangeListener("logLevel", new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent e) {
                String value = (String)e.getNewValue();
                int logLevel = TopGMapConfig.parseLogLevel(value);
                log.setLevel(logLevel);
            }
        }, true);
    }
    
    
    private void extractConfig() {
        config.setAutoAccuracy(autoAccuracy);
        config.setPosAccuracy(posAccuracy);
        config.setPosAccuracyWarn(posAccuracyWarn);
        config.setTrackMaxIncrement(trackMaxIncrement);
        
        config.setIgnoreLimits(ignoreLimits);
        config.setMaxWaypoints(maxWaypoints);
        config.setMaxRoutes(maxRoutes);
        config.setMaxTracks(maxTracks);
        
        int logLevel = log.getLevel();
        config.setLogLevel(logLevel);
    }
    
    
    public void process(String inputFileName, String outputFileName) throws IOException {
        Reader      in;
        PrintWriter out;
        
        if (inputFileName != null) {
            in = new FileReader(inputFileName);
        } else {
            in = new InputStreamReader(System.in);
        }
        
        if (outputFileName != null) {
            out = new PrintWriter(new FileWriter(outputFileName));
        } else {
            out = new PrintWriter(new OutputStreamWriter(System.out));
        }
        
        process(in, out);
        
        if (inputFileName != null) {
            in.close();
        }
        
        if (outputFileName != null) {
            out.close();
        }
    }
    
    
    public void process(Reader input, PrintWriter output) throws IOException {
        // Read OVL file
        OvlReader ovlReader = new OvlReader(input);
        GpsElement gpsElement;
        
        while ((gpsElement = ovlReader.nextElement()) != null) {
        }

        List allWpList = GpsWaypoint.instances();
        //GpsWaypoint[] gpsWaypoints = new GpsWaypoint[gpsWaypointList.size()];
        //gpsWaypoints = (GpsWaypoint[])gpsWaypointList.toArray(gpsWaypoints);
        
        log.debug("allWpList.size="   + allWpList.size());
        log.debug("allLineList.size=" + GpsLine.instances().size());
        
        if (autoAccuracy) {
            autoAccuracy();
        }

        // Check for duplicate line waypoints
        log.info("berprfen auf doppelte Linien-Wegpunkte.");
        for (Iterator iterLines = GpsLine.instances().iterator(); iterLines.hasNext(); ) {
            GpsLine line = (GpsLine)iterLines.next();
            
            List lineWpList = line.getWaypoints();
            
            GpsWaypoint lineWp1;
            GpsWaypoint lineWp2 = null;
            
            for (ListIterator iterLineWp = lineWpList.listIterator(); iterLineWp.hasNext(); ) {
                lineWp1 = lineWp2;
                
                lineWp2 = (GpsWaypoint)iterLineWp.next();
                
                if (lineWp1 == null) {
                    continue;
                }
                
                Point2D posLineWp1 = lineWp1.getPosition();
                Point2D posLineWp2 = lineWp2.getPosition();
                    
                double dist = GisUtils.dist(posLineWp1, posLineWp2);
                if (dist < posAccuracy || dist < posAccuracyWarn) {
                    log.warn("Doppelter Linien-Wegpunkt?",
                             new Object[] {
                                "Distance",   new Double(dist),
                                "Waypoint 1", lineWp1,
                                "Waypoint 2", lineWp2,
                    });
                }
                
            }
        }
        
        // Remove duplicate waypoints with equal accuracy, execpt unnamed waypoints
        log.info("Entfernen doppelter Wegpunkte.");
        for (Iterator iterAllWp = allWpList.iterator(); iterAllWp.hasNext(); ) {
            GpsWaypoint anyWp = (GpsWaypoint)iterAllWp.next();
            
            if (!anyWp.isWaypoint()) {
                continue;
            }
            
            if (anyWp.isUnnamed()) {
                continue;
            }
            
            Point2D posGpsAnyWp = anyWp.getPosition();
            
            for (Iterator iterDupWp = allWpList.iterator(); iterDupWp.hasNext(); ) {
                GpsWaypoint dupWp = (GpsWaypoint)iterDupWp.next();
                
                if (!dupWp.isWaypoint()) {
                    continue;
                }
                
                if (dupWp == anyWp) {
                    continue;
                }
                
                if (dupWp.getAccuracy() != anyWp.getAccuracy()) {
                    continue;
                }
                
                Point2D posDupWaypoint = dupWp.getPosition();
                
                double dist = GisUtils.dist(posGpsAnyWp, posDupWaypoint);
                if (dist < posAccuracy) {
                    dupWp.delete();
                    log.warn("Wegpunkt mit identischer Position entfernt!",
                             new Object[] {
                                "Distance",   new Double(dist),
                                "Waypoint 1", dupWp,
                                "Waypoint 2", anyWp,
                    });
                    break;
                } else
                if (dist < posAccuracyWarn) {
                    log.warn("Wegpunkt mit identischer Position?",
                             new Object[] {
                                "Distance",   new Double(dist),
                                "Waypoint 1", dupWp,
                                "Waypoint 2", anyWp,
                    });
                }
            }
            
        }
    
        // Replace reference waypoints by named waypoints
        log.info("Ersetzen von Referenz-Wegpunkten durch benannte Wegpunkte.");
        for (ListIterator iterNamedWp = allWpList.listIterator(); iterNamedWp.hasNext(); ) {
            GpsWaypoint namedWp = (GpsWaypoint)iterNamedWp.next();
            
            if (!namedWp.isWaypoint()) {
                continue;
            }
            
            if (!namedWp.isNamed()) {
                continue;
            }
            
            Point2D posNamedWp = namedWp.getPosition();
            
            for (ListIterator iterRefWp = allWpList.listIterator(); iterRefWp.hasNext(); ) {
                GpsWaypoint refWp = (GpsWaypoint)iterRefWp.next();
                
                if (!refWp.isWaypoint()) {
                    continue;
                }
                
                if (!refWp.isReference()) {
                    continue;
                }
                
                Point2D posRefWp = refWp.getPosition();
                
                double dist = GisUtils.dist(posNamedWp, posRefWp);
                if (dist < posAccuracy) {
                    log.debug("Waypoint position adopted.",
                             new Object[] {
                                "Distance",   new Double(dist),
                                "Waypoint 1", refWp,
                                "Waypoint 2", namedWp,
                    });

                    namedWp.setPosition(posRefWp);
                    refWp.delete();
                    break;
                } else
                if (dist < posAccuracyWarn) {
                    log.warn("Wegpunkte mit identischer Position?",
                             new Object[] {
                                "Distance",   new Double(dist),
                                "Waypoint 1", namedWp,
                                "Waypoint 2", refWp,
                    });
                }
            }
            
        }
        
        // Replace line waypoints by other waypoints
        log.info("Ersetzen von Linien-Wegpunkten.");
        for (Iterator iterLines = GpsLine.instances().iterator(); iterLines.hasNext(); ) {
            GpsLine line = (GpsLine)iterLines.next();
            
            List lineWpList = line.getWaypoints();
            
            for (ListIterator iterLineWp = lineWpList.listIterator(); iterLineWp.hasNext(); ) {
                GpsWaypoint lineWp = (GpsWaypoint)iterLineWp.next();
                
                assert lineWp.isUnnamed();
                
                Point2D posLineWp = lineWp.getPosition();
                
                double dist = Double.MAX_VALUE;
                GpsWaypoint replWp = lineWp;
                
                for (Iterator iterRefWp = allWpList.iterator(); iterRefWp.hasNext(); ) {
                    GpsWaypoint refWp = (GpsWaypoint)iterRefWp.next();
                    
                    if (!refWp.isWaypoint()) {
                        continue;
                    }
                    
                    if (refWp == lineWp) {
                        continue;
                    }
                    
                    Point2D posRefWp = refWp.getPosition();
                    
                    dist = GisUtils.dist(posLineWp, posRefWp);
                    if (dist < posAccuracy) {
                        if (refWp.getAccuracy() > replWp.getAccuracy()) {
                            replWp = refWp;
                            break;
                        }
                    } else
                    if (dist < posAccuracyWarn) {
                        log.warn("Wegpunkte mit identischer Position?",
                                 new Object[] {
                                    "Distance",   new Double(dist),
                                    "Waypoint 1", lineWp,
                                    "Waypoint 2", refWp,
                        });
                    }
                }
                
                if (replWp != lineWp) {                    
                    log.debug("Waypoint replaced (Wp1 by Wp2).",
                             new Object[] {
                                "Distance",   new Double(dist),
                                "Waypoint 1", lineWp,
                                "Waypoint 2", replWp,
                    });

                    iterLineWp.set(replWp);
                    
                    lineWp.delete();
                }
            }
        }
        
        // Replace waypoints with shared position of all lines
        log.info("Ersetzen von Wegpunkten mit identischer Position von allen Linien.");
        for (Iterator iterLines = GpsLine.instances().iterator(); iterLines.hasNext(); ) {
            GpsLine gpsLine = (GpsLine)iterLines.next();
            
            List lineWpList = gpsLine.getWaypoints();
            
            for (int indexLineWp1 = 0; indexLineWp1 < lineWpList.size() - 1; indexLineWp1++) {
                GpsWaypoint lineWp1 = (GpsWaypoint)lineWpList.get(indexLineWp1);
                
                Point2D posLineWp1 = lineWp1.getPosition();
                
                for (int indexLineWp2 = indexLineWp1 + 1; indexLineWp2 < lineWpList.size(); indexLineWp2++) {
                    GpsWaypoint lineWp2 = (GpsWaypoint)lineWpList.get(indexLineWp2);
                    
                    Point2D posLineWp2 = lineWp2.getPosition();

                    double dist = GisUtils.dist(posLineWp1, posLineWp2);
                    if (dist < posAccuracy) {
                        log.debug("Waypoint replaced (Wp2 by Wp1).",
                        new Object[] {
                            "Distance",   new Double(dist),
                            "Waypoint 1", lineWp1,
                            "Waypoint 2", lineWp2,
                        });
                        lineWpList.set(indexLineWp2, lineWp1);
                    } else
                    if (dist < posAccuracyWarn) {
                        log.warn("Wegpunkte mit identischer Position?",
                                 new Object[] {
                                     "Distance",   new Double(dist),
                                     "Waypoint 1", lineWp1,
                                     "Waypoint 2", lineWp2,
                        });
                    }
                }
            }
        }
        
        // Replace line waypoints with shared position
        log.info("Ersetze Linien-Wegpunkte mit identischer Position.");
        for (Iterator iterLines1 = GpsLine.instances().iterator(); iterLines1.hasNext(); ) {
            GpsLine gpsLine1 = (GpsLine)iterLines1.next();
            
            List lineWpList1 = gpsLine1.getWaypoints();
            
            for (Iterator iterLines2 = GpsLine.instances().iterator(); iterLines2.hasNext(); ) {
                GpsLine gpsLine2 = (GpsLine)iterLines2.next();
                
                List lineWpList2 = gpsLine2.getWaypoints();
                
                for (Iterator iterLineWp1 = lineWpList1.iterator(); iterLineWp1.hasNext(); ) {
                    GpsWaypoint lineWp1 = (GpsWaypoint)iterLineWp1.next();
                	
                    Point2D posLineWp1 = lineWp1.getPosition();
                    
                    for (ListIterator iterLineWp2 = lineWpList2.listIterator(); iterLineWp2.hasNext(); ) {
                    	GpsWaypoint lineWp2 = (GpsWaypoint)iterLineWp2.next();
                        
                        if (lineWp1 == lineWp2)
                        	continue;
                        
                        Point2D posLineWp2 = lineWp2.getPosition();

                        double dist = GisUtils.dist(posLineWp1, posLineWp2);
                        if (dist < posAccuracy) {
                        	if (gpsLine1 == gpsLine2) {
                                log.warn("Implementierungsbeschrnkung: Wegpunkte einer Linie mit identischer Position knnen nicht ersetzt werden.",
                                        new Object[] {
                                           "Distance",   new Double(dist),
                                           "Waypoint 1", lineWp1,
                                           "Waypoint 2", lineWp2,
                                });
                                continue;
                        	}
                            log.debug("Waypoint replaced (Wp2 by Wp1).",
                                    new Object[] {
                                       "Distance",   new Double(dist),
                                       "Waypoint 1", lineWp1,
                                       "Waypoint 2", lineWp2,
                            });
                        	iterLineWp2.set(lineWp1);
                        } else
                        if (dist < posAccuracyWarn) {
                            log.warn("Wegpunkte mit identischer Position?",
                                     new Object[] {
                                        "Distance",   new Double(dist),
                                        "Waypoint 1", lineWp1,
                                        "Waypoint 2", lineWp2,
                            });
                        }
                    
                    }
                }
            }
        }
        
        // Double link line waypoints
        log.info("Erzeuge fr Linien-Wegpunkte Verweise auf alle Nachbarpunkte.");
        for (Iterator iterLines = GpsLine.instances().iterator(); iterLines.hasNext(); ) {
            GpsLine gpsLine = (GpsLine)iterLines.next();
            
            List lineWpList = gpsLine.getWaypoints();
            
            GpsWaypoint lineWp2 = null;
            
            for (Iterator iterLineWp = lineWpList.iterator(); iterLineWp.hasNext(); ) {
                
                GpsWaypoint lineWp1 = lineWp2;
                    
                lineWp2 = (GpsWaypoint)iterLineWp.next();
                
                if (lineWp1 == null || lineWp2 == null) {
                    continue;
                }
                
                lineWp1.addNeighbour(lineWp2);
                lineWp2.addNeighbour(lineWp1);
            }
        }
                    
        // Create labeled routes and tracks and assign lines
        log.info("Erzeuge benannte Routen und Tracks und weise Linien zu.");
        for (Iterator iterLabelWp = allWpList.iterator(); iterLabelWp.hasNext(); ) {
            GpsWaypoint labelWp = (GpsWaypoint)iterLabelWp.next();
            
            if (!(labelWp.isRouteLabel() || labelWp.isTrackLabel())) {
                // No route or track name
                continue;
            }
            
            // Find nearest line to label
            Point2D posLabel = labelWp.getPosition();
            
            double minDist = Double.MAX_VALUE;
            GpsLine nearestLine = null;
            for (Iterator iterLine = GpsLine.instances().iterator(); iterLine.hasNext(); ) {
                GpsLine gpsLine = (GpsLine)iterLine.next();
                
                List lineWaypoints = gpsLine.getWaypoints();
                
                Point2D posLineWp1;
                Point2D posLineWp2 = null;
                for (Iterator iterLineWp = lineWaypoints.iterator(); iterLineWp.hasNext(); ) {
                    posLineWp1 = posLineWp2;
                    GpsWaypoint gpsWaypoint = (GpsWaypoint)iterLineWp.next();
                    posLineWp2 = gpsWaypoint.getPosition();
                    if (posLineWp1 == null) {
                        continue;
                    }
                    Point2D pp = Geometry.perpendicular(posLineWp1, posLineWp2, posLabel);
                    double dist = GisUtils.dist(pp, posLabel);
                    if (dist < minDist) {
                        minDist = dist;
                        nearestLine = gpsLine;
                    }
                }
            }
            
            if (nearestLine == null) {
                log.warn("Keine Linie fr Bezeichner gefunden.",
                         new Object[] {
                            "Waypoint 1", labelWp,
                });
                continue;
            }
            
            if (minDist > posAccuracy) {
                log.warn("Bezeichner an korrekter Position?",
                         new Object[] {
                            "Distance", new Double(minDist),
                            "Waypoint 1", labelWp,
                });
            }
            
            if (labelWp.isRouteLabel()) {
                String name = labelWp.getName();
                GpsRoute gpsRoute = GpsRoute.newInstance(name);
                gpsRoute.setGpsLine(nearestLine);
                log.debug("Labeled route created.",
                         new Object[] {
                            "Distance", new Double(minDist),
                            "Route",    gpsRoute,
                });
            } else
            if (labelWp.isTrackLabel()) {
                String name = labelWp.getName();
                GpsTrack gpsTrack = GpsTrack.newInstance(name);
                gpsTrack.setGpsLine(nearestLine);
                log.debug("Labeled track created.",
                         new Object[] {
                            "Distance", new Double(minDist),
                            "Track",    gpsTrack,
                });
            }
        }
        
        log.debug("gpsRoutes.size=" + GpsRoute.instances().size());
        log.debug("gpsTracks.size=" + GpsTrack.instances().size());
        
        // Abschtzung der Toleranz
        //   Top50: 1:50 000, 1 cm = 500 m; Auflsung 100 L/cm -> 1 Pixel = 5 m
        
        // Alternatives Vorgehen:
        //   Liste von allen Verbindungsmglichkeiten von Text (Route/Track) - Linie erzeugen.
        //   Eine Eigenschaft eines solchen Verbindungselements ist die Entfernung Text - Linie.

        // Alternatives Vorgehen:
        //   Fr jede Linie wird ein Track (zunchst ohne Namen) erzeugt.

        /*
        // Generate tracks with generic names for all other lines
        log.info("Generate tracks with generic names for all other lines.");
        for (Iterator iterLine = GpsLine.instances().iterator(); iterLine.hasNext(); ) {
            GpsLine gpsLine = (GpsLine)iterLine.next();
            
            boolean isUsed = false;
            for (Iterator iterTrack = GpsTrack.instances().iterator(); !isUsed && iterTrack.hasNext(); ) {
                GpsTrack gpsTrack = (GpsTrack)iterTrack.next();
                isUsed = gpsTrack.getGpsLine().equals(gpsLine);
            }
            
            if (!isUsed) {
                GpsTrack gpsTrack = GpsTrack.newInstance();
                gpsTrack.setGpsLine(gpsLine);
            }
        }
        
        for (Iterator iterTrack = GpsTrack.instances().iterator(); iterTrack.hasNext(); ) {
            GpsTrack gpsTrack = (GpsTrack)iterTrack.next();
            log.debug(gpsTrack.toString());
        }
        */
        
        if (!ignoreLimits) {
            // Observe the device limits
            log.info("Beachte Gerte-Beschrnkungen.");
            
            // Limit up to 10 saved tracks, one reserved for basemap
            log.info("Begrenze auf " + maxTracks + " gespeicherte Tracks und reserviere einen fr die Basiskarte.");
            List gpsTracks = GpsTrack.instances();
            while (gpsTracks.size() > maxTracks - 1) {
                GpsTrack gpsTrack = (GpsTrack)gpsTracks.get(gpsTracks.size() - 1);
                gpsTrack.delete();
                log.warn("Track gelscht!",
                         new Object[] {
                            "Track", gpsTrack,
                });
            }
        }
        
        // Create basemap track (traveling salesman problem, TSP)
        log.info("Erzeuge Basiskarten-Track.");
        createBasemap();
        
        if (!ignoreLimits) {
            // Observe the device limits
            log.info("Beachte Gerte-Beschrnkungen.");
            
            // Limit up to 500 waypoints
            log.info("Begrenze auf " + maxWaypoints + " Wegpunkte.");
            // @todo Remove only unnamed waypoints
            int nWp = 0;
            for (Iterator iterAllWp = allWpList.iterator(); iterAllWp.hasNext(); ) {
                GpsWaypoint anyWp = (GpsWaypoint)iterAllWp.next();
                if (anyWp.isDeleted())
                    continue;
                if (anyWp.isRouteLabel() || anyWp.isTrackLabel())
                    continue;
                if (anyWp.isUnnamed())
                    continue;
                nWp++;
                if (nWp > maxWaypoints) {
                    anyWp.delete();
                    log.warn("Wegpunkt gelscht!",
                             new Object[] {
                                "Waypoint 1", anyWp,
                    });
                }
            }
            
            // Limit up to 20 routes
            log.info("Begrenze auf " + maxRoutes + " Routen.");
            List gpsRoutes = GpsRoute.instances();
            while (gpsRoutes.size() > maxRoutes) {
                GpsRoute gpsRoute = (GpsRoute)gpsRoutes.get(gpsRoutes.size() - 1);
                gpsRoute.delete();
                log.warn("Route gelscht!",
                         new Object[] {
                            "Route", gpsRoute,
                });
            }
        }
        
        // First writing phase: generate GPS names only
        log.debug("Write preparation phase.");
        GpxWriter nullWriter = new GpxWriter(new PrintWriter(new NullOutputStream()));
        
        for (Iterator iterWaypoint = allWpList.iterator(); iterWaypoint.hasNext(); ) {
            GpsWaypoint gpsWaypoint = (GpsWaypoint)iterWaypoint.next();
            
            if (gpsWaypoint.isWaypoint() && gpsWaypoint.isNamed()) {
                nullWriter.write(gpsWaypoint);
            }
        }
        
        for (Iterator iterRoute = GpsRoute.instances().iterator(); iterRoute.hasNext(); ) {
            GpsRoute gpsRoute = (GpsRoute)iterRoute.next();
            // @todo Limit route up to a maximum of 125 waypoints
            nullWriter.write(gpsRoute);
        }
        
        for (Iterator iterTrack = GpsTrack.instances().iterator(); iterTrack.hasNext(); ) {
            GpsTrack gpsTrack = (GpsTrack)iterTrack.next();
            nullWriter.write(gpsTrack);
        }
        
        nullWriter.close();
        
        GpsWaypoint.reset();
        GpsRoute.reset();
        GpsTrack.reset();
        
        // Save as GPX file
        log.debug("Save as GPX file.");
        GpxWriter gpxWriter = new GpxWriter(output);
        
        for (Iterator iterWaypoint = allWpList.iterator(); iterWaypoint.hasNext(); ) {
            GpsWaypoint gpsWaypoint = (GpsWaypoint)iterWaypoint.next();
            
            if (gpsWaypoint.isWaypoint() && gpsWaypoint.isNamed()) {
                gpxWriter.write(gpsWaypoint);
            }
        }
        
        for (Iterator iterRoute = GpsRoute.instances().iterator(); iterRoute.hasNext(); ) {
            GpsRoute gpsRoute = (GpsRoute)iterRoute.next();
            // @todo Limit route up to a maximum of 125 waypoints
            gpxWriter.write(gpsRoute);
        }
        
        for (Iterator iterTrack = GpsTrack.instances().iterator(); iterTrack.hasNext(); ) {
            GpsTrack gpsTrack = (GpsTrack)iterTrack.next();
            gpxWriter.write(gpsTrack);
        }
        
        gpxWriter.close();
    }
    
    
    static class Dist implements Comparable {
        GpsWaypoint wp1;
        GpsWaypoint wp2;
        double dist;
        
        public Dist(GpsWaypoint wp1, GpsWaypoint wp2) {
            this.wp1 = wp1;
            this.wp2 = wp2;
            calcDist();
        }
        
        private void calcDist() {
            Point2D posWp1 = wp1.getPosition();
            Point2D posWp2 = wp2.getPosition();
            dist = GisUtils.dist(posWp1, posWp2);
        }
        
        public double getDist() {
            return dist;
        }
        
        public int compareTo(Object o) {
            double diff = dist - ((Dist)o).dist;
            return diff < 0 ? -1 : diff > 0 ? 1 : 0;
        }
        
        public String toString() {
            StringBuffer buffer = new StringBuffer("Dist@");
            buffer.append(Integer.toHexString(hashCode()));
            buffer.append("[wp1=");
            buffer.append(wp1);
            buffer.append(",wp2=");
            buffer.append(wp2);
            buffer.append(",dist=");
            buffer.append(dist);
            buffer.append("]");
            return buffer.toString();
        }
        
    }
    
    
    public void autoAccuracy() {
        log.info("Genauigkeit automatisch ermitteln.");
        
        // Ermittle minimalen Abstand zwischen zwei aufeinanderfolgender Wegpunkte.
        double minDist = Double.MAX_VALUE;
        GpsWaypoint minWp1 = null;
        GpsWaypoint minWp2 = null;
        for (Iterator iterLines = GpsLine.instances().iterator(); iterLines.hasNext(); ) {
            GpsLine line = (GpsLine)iterLines.next();
            
            List lineWpList = line.getWaypoints();
            
            GpsWaypoint lineWp1;
            GpsWaypoint lineWp2 = null;
            
            for (ListIterator iterLineWp = lineWpList.listIterator(); iterLineWp.hasNext(); ) {
                lineWp1 = lineWp2;
                
                lineWp2 = (GpsWaypoint)iterLineWp.next();
                
                if (lineWp1 == null) {
                    continue;
                }
                
                Point2D posLineWp1 = lineWp1.getPosition();
                Point2D posLineWp2 = lineWp2.getPosition();
                    
                double dist = GisUtils.dist(posLineWp1, posLineWp2);
                
                if (dist == 0.0) {
                    log.warn("Doppelten Wegpunkt ignoriert.",
                            new Object[] {
                                "Distance", new Double(dist),
                                "Waypoint 1", lineWp1,
                                "Waypoint 2", lineWp2,
                    });
                    continue;
                }
                
                if (dist < minDist) {
                    minDist = dist;
                    minWp1 = lineWp1;
                    minWp2 = lineWp2;
                }
            }
        }
        
        // Annherung ist minimale Entfernung zweier aufeinanderfolgender Wegpunkte.
        posAccuracyWarn = minDist;
        posAccuracyWarn = Math.floor(minDist * 10d) / 10d;

        log.info("Ermittelte Annherung.",
                new Object[] {
                    "Distance", new Double(posAccuracyWarn),
                    "Waypoint 1", minWp1,
                    "Waypoint 2", minWp2,
        });
        
        // Annahme: Genauigkeit besser als halbe Annherung.
        posAccuracy = posAccuracyWarn / 2;
        
        // Ermittle maximalen Abstand zweier Wegpunkte, der kleiner als angenommene Genauigkeit ist.
        List allWpList = GpsWaypoint.instances();
        double maxDist = 0;
        GpsWaypoint maxWp1 = null;
        GpsWaypoint maxWp2 = null;
        for (int i = 0; i < allWpList.size() - 1; i++) {
            GpsWaypoint wp1 = (GpsWaypoint)allWpList.get(i);
            for (int k = i + 1; k < allWpList.size(); k++) {
                GpsWaypoint wp2 = (GpsWaypoint)allWpList.get(k);
                
                Point2D posWp1 = wp1.getPosition();
                Point2D posWp2 = wp2.getPosition();
                    
                double dist = GisUtils.dist(posWp1, posWp2);
                if (dist < posAccuracy && dist > maxDist) {
                    maxDist = dist;
                    maxWp1 = wp1;
                    maxWp2 = wp2;
                }
            }
        }
        
        // Genauigkeit ist maximaler Abstand.
        posAccuracy     = maxDist;
        posAccuracy     = Math.ceil(posAccuracy * 10d) / 10d;
        
        log.info("Ermittelte Genauigkeit.",
                new Object[] {
                    "Distance", new Double(maxDist),
                    "Waypoint 1", maxWp1,
                    "Waypoint 2", maxWp2,
        });
        
        //throw new RuntimeException("DIESE EXCEPTION IST NUR EIN TEST");
    }
    
    
    private GpsLine basemapLine  = null;
    private boolean legVisited[][];
    
    
    // Create basemap track (traveling salesman problem, TSP)
    public void createBasemap() {    
        
        int nWaypoints = GpsWaypoint.nextId();
        legVisited = new boolean[nWaypoints][nWaypoints];
        for (int i = 0; i < nWaypoints; i++) {
            legVisited[i] = new boolean[nWaypoints];
            for (int k = 0; k < nWaypoints; k++) {
                legVisited[i][k] = false;
            }
        }

        GpsWaypoint rootWaypoint = findRootWaypoint();
        if (rootWaypoint == null) {
            return;
        }
        log.debug("Root Waypoint",
                 new Object[] {
                    "Waypoint 1",    rootWaypoint,
        });
        
        addBasemapWaypoint(rootWaypoint);
        
        crossWaypoint(rootWaypoint);
        
        GpsTrack basemapTrack = GpsTrack.newInstance("ACTIVE LOG", basemapLine);
        
        // Check whether all lines where visited 
        log.info("berprfe auf unzusammenhngende Linien.");
        for (Iterator iterLines = GpsLine.instances().iterator(); iterLines.hasNext(); ) {
            GpsLine line = (GpsLine)iterLines.next();
            
            List lineWpList = line.getWaypoints();
            
            GpsWaypoint lineWp1;
            GpsWaypoint lineWp2 = null;
            boolean visited = false;
            
            for (ListIterator iterLineWp = lineWpList.listIterator(); iterLineWp.hasNext(); ) {
                lineWp1 = lineWp2;
                lineWp2 = (GpsWaypoint)iterLineWp.next();
                
                if (lineWp1 == null) {
                    continue;
                }
                
                visited |= isLegVisited(lineWp1, lineWp2);
                
                if (visited)
                    break;
            }
            
            if (!visited) {
                log.warn("Unverbundene Linie von Basiskarte weggelassen.!",
                        new Object[] {
                           "Line", line,
               });
            }
        }
    }
    
    
    private GpsWaypoint findRootWaypoint() {
        for (Iterator iterLine = GpsLine.instances().iterator(); iterLine.hasNext(); ) {
            GpsLine gpsLine = (GpsLine)iterLine.next();
            
            List gpsWaypoints = gpsLine.getWaypoints();
            for (Iterator iterWaypoints = gpsWaypoints.iterator(); iterWaypoints.hasNext(); ) {
                GpsWaypoint rootWaypoint = (GpsWaypoint)iterWaypoints.next();
                
                if (rootWaypoint != null) {
                    return rootWaypoint;
                }
            }
        }
        
        return null;
    }
    
    
    private void crossWaypoint(GpsWaypoint gpsWaypoint) {
        for (Iterator iterNeighbours = gpsWaypoint.neighbours().iterator(); iterNeighbours.hasNext(); ) {
            GpsWaypoint neighbourWaypoint = (GpsWaypoint)iterNeighbours.next();
            
            if (isLegVisited(gpsWaypoint, neighbourWaypoint)) {
                continue;
            }
            
            crossLeg(gpsWaypoint, neighbourWaypoint);
            crossWaypoint(neighbourWaypoint);
            crossLeg(neighbourWaypoint, gpsWaypoint);
        }
    }
    
    
    private void crossLeg(GpsWaypoint srcWaypoint, GpsWaypoint dstWaypoint) {
        addLegVisited(srcWaypoint, dstWaypoint);
        
        // Zwischenschritte errechnen und der Basemap hinzufgen
        Point2D srcPos = srcWaypoint.getPosition();
        Point2D dstPos = dstWaypoint.getPosition();
        double dist = GisUtils.dist(srcPos, dstPos);
        int nSteps = trackMaxIncrement > 0 ? (int)Math.ceil(dist / trackMaxIncrement) : 1;
        
        double latBase = srcPos.getX();
        double lonBase = srcPos.getY();
        double latStep = (dstPos.getX() - srcPos.getX()) / nSteps;
        double lonStep = (dstPos.getY() - srcPos.getY()) / nSteps;
        for (int iStep = 1; iStep < nSteps; iStep++) {
            double lat = latBase + iStep * latStep;
            double lon = lonBase + iStep * lonStep;
            Point2D position = new Point2D.Double(lat, lon);
            GpsWaypoint stepWaypoint = GpsWaypoint.newInstance(position);
            addBasemapWaypoint(stepWaypoint);
        }
        
        // Letzter Schritt zum Zielpunkt
        addBasemapWaypoint(dstWaypoint);
    }
    
    
    private boolean isLegVisited(GpsWaypoint srcWaypoint, GpsWaypoint dstWaypoint) {
        int wpId1 = srcWaypoint.getId();
        int wpId2 = dstWaypoint.getId();
        if (wpId1 < wpId2) {
            int wpIdTmp = wpId1;
            wpId1 = wpId2;
            wpId2 = wpIdTmp;
        }
        
        return legVisited[wpId1][wpId2];
    }
    
    
    private void addLegVisited(GpsWaypoint srcWaypoint, GpsWaypoint dstWaypoint) {
        if (srcWaypoint == null) {
            return;
        }
        int wpId1 = srcWaypoint.getId();
        int wpId2 = dstWaypoint.getId();
        if (wpId1 < wpId2) {
            int wpIdTmp = wpId1;
            wpId1 = wpId2;
            wpId2 = wpIdTmp;
        }
        
        legVisited[wpId1][wpId2] = true;
    }
    
    
    private void addBasemapWaypoint(GpsWaypoint gpsWaypoint) {
        if (basemapLine == null) {
            basemapLine = GpsLine.newInstance();
        }
        log.debug("Basemap extended.",
                 new Object[] {
                    "Waypoint 1",    gpsWaypoint,
        });
        basemapLine.add(gpsWaypoint);
    }
    
    
    public static String outputFileName(String inputFileName) {
        String nameExt = FileUtils.nameExt(inputFileName);
        String coreName = "ovl".equalsIgnoreCase(nameExt) ? FileUtils.coreName(inputFileName) : inputFileName;
        String outputFileName = coreName + ".gpx";
        
        return outputFileName;    
    }
    
    
    public static void main(String[] args) {
        String inputFileName  = null;
        String outputFileName = null;
        
        int state = 0;
        int iArgs = 0;
        String arg;
        while (state != 4) {
            switch (state) {
            case 0:
                if (iArgs == args.length) {
                    state = 4;
                    continue;
                }
                arg = args[iArgs++];
                if (arg.equals("-i")) {
                    state = 1;
                } else
                if (arg.equals("-o")) {
                    state = 2;
                } else
                if (arg.equals("-h")) {
                    state = 3;
                }
                break;
            case 1:
                if (iArgs == args.length) {
                    log.error("Missing Argument: Input File");
                    System.exit(1);
                }
                arg = args[iArgs++];
                inputFileName = arg;
                state = 0;
                break;
            case 2:
                if (iArgs == args.length) {
                    log.error("Missing Argument: Output File");
                    System.exit(1);
                }
                arg = args[iArgs++];
                outputFileName = arg;
                state = 0;
                break;
            case 3:
        	log.error(
        		"USAGE: TopGMap [OPTION]...\n" +
        		"Options:\n" +
        		"  -h		print a synopsis of options\n" +
        		"  -i FILE	specification of input file\n" +
        		"  -o FILE	specification of output file");
        	System.exit(1);
                state = 0;
                break;
            default:
                log.fatal("Internal error: Illegal state");
    		System.exit(1);
                break;
            }
        }
        
        TopGMap topGMap = new TopGMap();
        try {
            topGMap.process(inputFileName, outputFileName);
            topGMap.close();
        } catch (Exception ex) {
            log.fatal("IO failed: " + ex);
	    System.exit(1);
        }
	System.exit(0);
    }
    
}
