/*
 * TopGMapConfig.java
 *
 * Copyright (c) 2005 Bernhard Fischer.
 * This file is part of TopGMap. TopGMap is free software; it is
 * published under the terms of the GNU General Public License.
 */

package de.bfisch.gis;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.beans.PropertyChangeListener;

import javax.swing.filechooser.FileSystemView;

import de.bfisch.util.BeanProperties;
import de.bfisch.util.LogManager;
import de.bfisch.util.Logger;


public class TopGMapConfig {
        
    private static final Logger log = new Logger("TopGMapConfig");
    
    private static final String CONFIG_FILE_NAME = ".TopGMap.properties";
    
    private static TopGMapConfig theInstance;
    
    private String name;
    private BeanProperties props;
    
    
    public TopGMapConfig(String name) {
        this.name  = name;
        this.props = new BeanProperties(System.getProperties());
    }
    
    
    public static TopGMapConfig getInstance(String name) {
        if (theInstance == null) {
            theInstance = new TopGMapConfig(name);
        }
        return theInstance;
    }
    
    
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        props.addPropertyChangeListener(listener);
    }
    
    
    public void addPropertyChangeListener(PropertyChangeListener listener, boolean init) {
        props.addPropertyChangeListener(listener);
        if (init) {
            props.firePropertyChange();
        }
    }
    
    
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        props.removePropertyChangeListener(listener);
    }
    
    
    public void removePropertyChangeListener(PropertyChangeListener listener, boolean init) {
        props.removePropertyChangeListener(listener);
        if (init) {
            props.firePropertyChange();
        }
    }
    
    
    public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener) {
        addPropertyChangeListener(propertyName, listener, false);
    }
    
    
    public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener, boolean init) {
        props.addPropertyChangeListener(propertyName, listener);
        if (init) {
            props.firePropertyChange(propertyName);
        }
    }
    
    
    public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener) {
        removePropertyChangeListener(propertyName, listener, false);
    }
    
    
    public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener, boolean init) {
        props.removePropertyChangeListener(propertyName, listener);
        if (init) {
            props.firePropertyChange(propertyName);
        }
    }
    
    
    public String getName() {
        return name;
    }
    
    
    public Object set(String key, String value) {
        return props.setProperty(key, value);
    }
    
    
    public String get(String key) {
        return props.getProperty(key);
    }
    
    
    public String get(String key, String defaultValue) {
        return props.getProperty(key, defaultValue);
    }
    
    
    private File getDefaultDirectory() {
	FileSystemView fileSystemView = FileSystemView.getFileSystemView();
	File defaultDir = fileSystemView.getDefaultDirectory();
	return defaultDir;
    }
    
    
    public void load() throws IOException {
	File defaultDir = getDefaultDirectory();
        File propsFile = new File(defaultDir, CONFIG_FILE_NAME);
        if (!propsFile.isFile() || !propsFile.canRead()) {
            return;
        }
        InputStream in = new FileInputStream(propsFile);
        props.load(in);
        in.close();
    }
    
    
    public void store() throws IOException {
        assert props != null;
	File defaultDir = getDefaultDirectory();
        File propsFile = new File(defaultDir, CONFIG_FILE_NAME);
        if (propsFile.isFile() && !propsFile.canWrite()) {
            log.error("Can't store configuration");
            return;
        }
        OutputStream out = new FileOutputStream(propsFile);
        props.store(out, "TopGMap Configuration");
        out.close();
    }
    
    
    public void setInputPath(String inputPath) {
        set("inputPath", inputPath);
    }
    
    
    public String getInputPath() {
        String inputPath = get("inputPath");
        return inputPath;
    }
    
    
    public void setOutputPath(String outputPath) {
        set("outputPath", outputPath);
    }
    
    
    public String getOutputPath() {
        String outputPath = get("outputPath");
        return outputPath;
    }
    
    
    public void setCurrentDir(File currentDir) {
        set("currentDir", currentDir.getPath());
    }
    
    
    public File getCurrentDir() {
        String currentDirPath = get("currentDir");
        File currentDir = new File(currentDirPath);
        return currentDir;
    }
    
    
    public void setX(int x) {
        set("x", Integer.toString(x));
    }


    public int getX(int defaultX) {
        String xStr = get("x");
        int x = defaultX;
        try {
            x = Integer.parseInt(xStr);
        } catch (NumberFormatException ex) { }
        return x;
    }
    
    
    public void setY(int y) {
        set("y", Integer.toString(y));
    }


    public int getY(int defaultY) {
        String yStr = get("y");
        int y = defaultY;
        try {
            y = Integer.parseInt(yStr);
        } catch (NumberFormatException ex) { }
        return y;
    }
    
    
    public void setAutoAccuracy(boolean autoAccuracy) {
        set("autoAccuracy", Boolean.toString(autoAccuracy));
    }
    
    
    public static boolean parseAutoAccuracy(String autoAccuracyStr) {
        boolean autoAccuracy = parseBoolean(autoAccuracyStr, true);
        return autoAccuracy;
    }
    
    
    public boolean getAutoAccuracy() {
        String autoAccuracyStr = get("autoAccuracy");
        boolean autoAccuracy = parseAutoAccuracy(autoAccuracyStr);
        return autoAccuracy;
    }
    
    
    public void setPosAccuracy(double posAccuracy) {
        set("posAccuracy", Double.toString(posAccuracy));
    }
    
    
    public static double parsePosAccuracy(String posAccuracyStr) {
        double posAccuracy = parseDouble(posAccuracyStr, 5d * 5d);      // 5 pixel * 5 Meter map resolution
        return posAccuracy;
    }

    
    public double getPosAccuracy() {
        String posAccuracyStr = get("posAccuracy");
        double posAccuracy = parsePosAccuracy(posAccuracyStr);
        return posAccuracy;
    }
    
    
    public void setPosAccuracyWarn(double posAccuracyWarn) {
        set("posAccuracyWarn", Double.toString(posAccuracyWarn));
    }
    
    
    public static double parsePosAccuracyWarn(String posAccuracyWarnStr) {
        double posAccuracyWarn = parseDouble(posAccuracyWarnStr, 2 * 5d * 5d);  // doppelte Positioniergenauigkeit
        return posAccuracyWarn;
    }


    public double getPosAccuracyWarn() {
        String posAccuracyWarnStr = get("posAccuracyWarn");
        double posAccuracyWarn = parsePosAccuracyWarn(posAccuracyWarnStr);
        return posAccuracyWarn;
    }
    
    
    public void setIgnoreLimits(boolean ignoreLimits) {
        set("ignoreLimits", Boolean.toString(ignoreLimits));
    }
    
    
    public static boolean parseIgnoreLimits(String ignoreLimitsStr) {
        boolean ignoreLimits = parseBoolean(ignoreLimitsStr, true);
        return ignoreLimits;
    }
    
    
    public boolean getIgnoreLimits() {
        String ignoreLimitsStr = get("ignoreLimits");
        boolean ignoreLimits = parseIgnoreLimits(ignoreLimitsStr);
        return ignoreLimits;
    }
    
    
    public void setTrackMaxIncrement(int trackMaxIncrement) {
        set("trackMaxIncrement", Integer.toString(trackMaxIncrement));
    }


    public int getTrackMaxIncrement() {
        String trackMaxIncrementStr = get("trackMaxIncrement");
        int trackMaxIncrement = parseTrackMaxIncrement(trackMaxIncrementStr);
        return trackMaxIncrement;
    }
    
    
    public static int parseTrackMaxIncrement(String trackMaxIncrementStr) {
        int trackMaxIncrement = parseInt(trackMaxIncrementStr, 5000);           // 5000 Meter
        return trackMaxIncrement;
    }
    
    
    public void setMaxWaypoints(int maxWaypoints) {
        set("maxWaypoints", Integer.toString(maxWaypoints));
    }
    
    
    public static int parseMaxWaypoints(String maxWaypointsStr) {
        int maxWaypoints = parseInt(maxWaypointsStr, 500);
        return maxWaypoints;
    }
    

    public int getMaxWaypoints() {
        String maxWaypointsStr = get("maxWaypoints");
        int maxWaypoints = parseMaxWaypoints(maxWaypointsStr);
        return maxWaypoints;
    }
    
    
    public void setMaxRoutes(int maxRoutes) {
        set("maxRoutes", Integer.toString(maxRoutes));
    }

    
    public static int parseMaxRoutes(String maxRoutesStr) {
        int maxRoutes = parseInt(maxRoutesStr, 20);
        return maxRoutes;
    }

    
    public int getMaxRoutes() {
        String maxRoutesStr = get("maxRoutes");
        int maxRoutes = parseMaxRoutes(maxRoutesStr);
        return maxRoutes;
    }
    
    
    public void setMaxTracks(int maxTracks) {
        set("maxTracks", Integer.toString(maxTracks));
    }

    
    public static int parseMaxTracks(String maxTracksStr) {
        int maxTracks = parseInt(maxTracksStr, 10);
        return maxTracks;
    }
    
    
    public int getMaxTracks() {
        String maxTracksStr = get("maxTracks");
        int maxTracks = parseMaxTracks(maxTracksStr);
        return maxTracks;
    }
    
    
    public void setLogLevel(int logLevel) {
        set("logLevel", Logger.getLevelAsString(logLevel));
    }

    
    public static int parseLogLevel(String logLevelStr) {
        int logLevel = LogManager.WARN;
        if (logLevelStr != null) {
            try {
                logLevel = Logger.parseLevel(logLevelStr);
            } catch (IllegalArgumentException ex) { }
        }
        return logLevel;
    }

    
    public int getLogLevel() {
        String logLevelString = get("logLevel");
        int logLevel = parseLogLevel(logLevelString);
        return logLevel;
    }
    
    
    private static boolean parseBoolean(String valueStr, boolean defaultValue) {
        boolean value = defaultValue;
        if (valueStr != null) {
            try {
                value = Boolean.parseBoolean(valueStr);
            } catch (NumberFormatException ex) { }
        }
        return value;
    }
    
    
    private static int parseInt(String valueStr, int defaultValue) {
        int value = defaultValue;
        if (valueStr != null) {
            try {
                value = Integer.parseInt(valueStr);
            } catch (NumberFormatException ex) { }
        }
        return value;
    }
    
    
    private static double parseDouble(String valueStr, double defaultValue) {
        double value = defaultValue;
        if (valueStr != null) {
            try {
                value = Double.parseDouble(valueStr);
            } catch (NumberFormatException ex) { }
        }
        return value;
    }

}
