/*
 * JLogFrame.java
 *
 * Copyright (c) 2005 Bernhard Fischer.
 * This file is part of TopGMap. TopGMap is free software; it is
 * published under the terms of the GNU General Public License.
 */

package de.bfisch.gis.gui;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.TransferHandler;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TableColumnModelEvent;
import javax.swing.event.TableColumnModelListener;
import javax.swing.filechooser.FileFilter;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;

import de.bfisch.gis.LogEntryFormatter;
import de.bfisch.gis.TopGMapConfig;
import de.bfisch.swing.GlobalCopyAction;
import de.bfisch.swing.TableFormatter;
import de.bfisch.swing.TablePopupMenuHandler;
import de.bfisch.swing.BasicTableCellRenderer;
import de.bfisch.swing.BasicTableTransferHandler;
import de.bfisch.swing.NameExtFileFilter;
import de.bfisch.swing.TableSorter;
import de.bfisch.util.LogEntry;
import de.bfisch.util.LogHandler;
import de.bfisch.util.Logger;


public class JLogFrame extends JFrame {


    public static class LogTableFormatter extends LogEntryFormatter implements TableFormatter {
        
        private boolean keyPrefix;
        
        
        public LogTableFormatter(boolean keyPrefix) {
            this.keyPrefix = keyPrefix;
        }
        
        
        public String valueToData(JTable table, Object value, int row, int col) {
            TableModel tableModel = table.getModel();
            // Consider column reordering. 
            int modelCol = table.getColumnModel().getColumn(col).getModelIndex();
            String key = tableModel.getColumnName(modelCol);
            String valueType = key;
            boolean keyPrefix = this.keyPrefix & col >= 3;
            return valueToData(key, value, valueType, keyPrefix);
        }
        
    }
    
    
    public class PropertyAction extends AbstractAction {
        
        
        public PropertyAction() {
            this("Properties");
        }
        
        
        public PropertyAction(String name) {
            super(name);
            logTable.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
                public void valueChanged(ListSelectionEvent e) {
                    int selectedRowCount = logTable.getSelectedRowCount();
                    setEnabled(selectedRowCount == 1);
                }
            });
        }
        
        
        public void actionPerformed(ActionEvent e) {
            int viewRow = logTable.getSelectedRow();
            if (viewRow < 0)
                return;
            
            int modelIndex = logTableModelSorted.modelIndex(viewRow);
            LogEntry logEntry = (LogEntry)logEntries.get(modelIndex);
            
            showLogEntryDialog(logEntry);
        }
        
    }
    
    
    private static final int COL_TIMESTAMP  = 0;
    private static final int COL_LEVEL	    = 1;
    private static final int COL_MESSAGE    = 2;
    
    private static final Logger log = new Logger("JLogFrame");
    private TopGMapConfig config;

    private JTable logTable;
    private AbstractTableModel logTableModel;
    private TableSorter logTableModelSorted;
    private JFileChooser fileChooser;
    
    private JLogEntryDialog logEntryDialog;
    
    private List logEntries   = new ArrayList();
    
    
    public JLogFrame() {
        super("Protokoll - TopGMap");
        init();
    }
    
    
    private void init() {
        config = TopGMapConfig.getInstance("JLogFrame");

        setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
        
        Container contentPane = getContentPane();
        
        logTableModel = new AbstractTableModel() {
            private String[] columnNames = { "Timestamp", "Level", "Message" };
            
            public String getColumnName(int col) {
        	return columnNames[col];
            }
            
            public int getColumnCount() {
                return columnNames.length;
            }
            
            public int getRowCount() {
                int rowCount = logEntries.size();
                return rowCount;
            }
            
            public Object getValueAt(int row, int col) {
                LogEntry logEntry = (LogEntry)logEntries.get(row);
                switch (col) {
                case COL_TIMESTAMP:
                    long timeStamp = logEntry.getTime();
                    return new Long(timeStamp);
                case COL_LEVEL:
                    int logLevel = logEntry.getLevel();
                    return new Integer(logLevel);
                case COL_MESSAGE:
                    String msg = logEntry.getMessage();
                    return msg;
                }
                assert false;
                return null;
            }
            
        };

        logTableModelSorted = new TableSorter(logTableModel);
        
        logTable = new JTable(logTableModelSorted);
        logTableModelSorted.setTableHeader(logTable.getTableHeader());
        logTableModelSorted.setTable(logTable);
        
        TableCellRenderer logTableRenderer = new BasicTableCellRenderer(new LogTableFormatter(false));
        logTable.setAutoResizeMode(JTable.AUTO_RESIZE_LAST_COLUMN);
        TableColumnModel tableColumnModel = logTable.getColumnModel();
        for (int colIdx = 0; colIdx < tableColumnModel.getColumnCount(); colIdx++) {
            TableColumn tableColumn = tableColumnModel.getColumn(colIdx);
            tableColumn.setCellRenderer(logTableRenderer);
            switch (colIdx) {
            case COL_TIMESTAMP:
        	tableColumn.setPreferredWidth(80);
        	tableColumn.setMaxWidth(80);
        	break;
            case COL_LEVEL:
        	tableColumn.setPreferredWidth(50);
        	tableColumn.setMaxWidth(50);
        	break;
            case COL_MESSAGE:
        	tableColumn.setPreferredWidth(200);
        	break;
            }
        }
        
        TransferHandler logTableTransferHandler =
                new BasicTableTransferHandler(BasicTableTransferHandler.FORMAT_ROW_OBJECT) {
            private LogEntryFormatter logEntryFormatter = new LogEntryFormatter();
            
            public String objectToData(JTable table, Object obj) {
                LogEntry logEntry = (LogEntry)obj;
                String valueStr = logEntryFormatter.objectToData(logEntry);
                return valueStr;
            }
            
            public Object getObjectAtRow(JTable table, int row) {
                TableModel tableModel = table.getModel();
                if (tableModel instanceof TableSorter) {
                    row = ((TableSorter)tableModel).modelIndex(row);
                }
                LogEntry logEntry = (LogEntry)logEntries.get(row);
                return logEntry;
            }
        };
        logTable.setTransferHandler(logTableTransferHandler);
        
        // Handle double click
        logTable.addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
                if (e.getClickCount() == 2) {
                    int viewRow = logTable.rowAtPoint(e.getPoint());
                    int modelIndex = logTableModelSorted.modelIndex(viewRow);
                    LogEntry logEntry = (LogEntry)logEntries.get(modelIndex);
                    
                    showLogEntryDialog(logEntry);
                 }
            }
        });
        
        // Attach popup menu to logTable
        JPopupMenu logTablePopupMenu = new JPopupMenu();
        logTable.addMouseListener(new TablePopupMenuHandler(logTablePopupMenu));
        
        JMenuItem copyMenuItem = new JMenuItem(new GlobalCopyAction("Kopieren"));
        logTablePopupMenu.add(copyMenuItem);
        
        JMenuItem propertiesMenuItem = new JMenuItem(new PropertyAction("Eigenschaften"));
        logTablePopupMenu.add(propertiesMenuItem);
        
        JScrollPane scrollPane = new JScrollPane(logTable);
        contentPane.add(scrollPane, BorderLayout.CENTER);
        
        Dimension size = scrollPane.getViewport().getPreferredSize();
        size.height = 15 * logTable.getRowHeight();
        logTable.setPreferredScrollableViewportSize(size);        
        
        LogHandler logHandler = new LogHandler() {
            
            public void handle(LogEntry logEntry) {
                logEntries.add(logEntry);
            }
            
        };
        log.setLogHandler(logHandler);
        
        JPanel controlPanel = new JPanel();
        controlPanel.setLayout(new FlowLayout());
        contentPane.add(controlPanel, BorderLayout.SOUTH);
        
        JPanel controlButtonPanel = new JPanel();
        controlButtonPanel.setLayout(new GridLayout(1, 0, 4, 0));
        controlPanel.add(controlButtonPanel);

        JButton saveButton = new JButton("Speichern...");
        controlButtonPanel.add(saveButton);

        saveButton.addActionListener(new ActionListener() {    
            public void actionPerformed(ActionEvent e) {
                saveLog();
            }
        });
        
        JButton closeButton = new JButton("Schlieen");
        controlButtonPanel.add(closeButton);

        closeButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                hideFrame();
            }
        });
        
        pack();
    }
    
    
    public AbstractTableModel getLogTableModel() {
        return logTableModel;
    }
    
    
    public LogEntry getLogEntry(int modelIndex) {
        return (LogEntry)logEntries.get(modelIndex);
    }
    
    
    private void showLogEntryDialog(LogEntry logEntry) {
	if (logEntryDialog == null) {
	    logEntryDialog = new JLogEntryDialog(this);
	}
        logEntryDialog.showDialog("Log-Eintrag", logEntry);
    }
    
    
    private void saveLog() {
        fileChooser = new JFileChooser();
        
        File currentDir = config.getCurrentDir();
        fileChooser.setCurrentDirectory(currentDir);
        fileChooser.setSelectedFile(new File("TopGMap.log"));
        
        FileFilter logFileFilter = new NameExtFileFilter("LOG");
        fileChooser.setFileFilter(logFileFilter);
        
        int returnVal = fileChooser.showSaveDialog(JLogFrame.this);

        if (returnVal == JFileChooser.APPROVE_OPTION) {
            File logFile = fileChooser.getSelectedFile();
            
            if (logFile.exists()) {
                int option = JOptionPane.showConfirmDialog(JLogFrame.this,
                        "Die Datei " + logFile.getName() + " besteht bereits.\nMchten Sie diese berschreiben?",
                        "Speichern unter",
                        JOptionPane.YES_NO_OPTION);
                if (option != JOptionPane.YES_OPTION) {
                    return;
                }
            }
            
            try {
                PrintWriter out = new PrintWriter(logFile);
                
                int rowCount = logTableModel.getRowCount();
                int colCount = logTableModel.getColumnCount();

                String hdr1 = logTableModel.getColumnName(COL_TIMESTAMP).toUpperCase();
                String hdr2 = logTableModel.getColumnName(COL_LEVEL).toUpperCase();
                String hdr3 = logTableModel.getColumnName(COL_MESSAGE).toUpperCase();
                
                out.printf("%-12.12s %-5.5s %s", new Object[] { hdr1, hdr2, hdr3 });
                out.println();
                
                LogTableFormatter logTableFormatter = new LogTableFormatter(true);
                for (int row = 0; row < rowCount; row++) {
                    LogEntry logEntry = (LogEntry)logEntries.get(row);
                    
                    Object[] props = logEntry.properties();
                    for (int idxProps = 0; idxProps < props.length; idxProps += 2) {
                        Object key       = props[idxProps];
                        Object value     = props[idxProps + 1];
                        String valueType = (String)key;
                        String valueStr = logTableFormatter.valueToData(key, value, valueType, idxProps / 2 > 2);
                        // Split lines: DOS/Windows/OS/2: "\r\n"; Apple: "\r"; UNIX: "\n"
                        String[] valueLines = valueStr.split("\r\n|\r|\n");
                        if (valueLines.length > 1) {
                            for (int i = 0; i < valueLines.length; i++) {
                                out.println();
                                out.print("                   ");
                                out.print(valueLines[i]);
                            }
                        } else {
                            switch (idxProps / 2) {
                            case COL_TIMESTAMP:
                                out.printf("%-12.12s", valueStr);
                                break;
                            case COL_LEVEL:
                                out.printf(" %-5.5s", valueStr);
                                break;
                            case COL_MESSAGE:
                                out.print(" ");
                                out.print(valueStr);
                                break;
                            }
                        }
                    }
                    out.println();
                }
                
                out.close();
            } catch (FileNotFoundException ex) {
                ex.printStackTrace();
            }
        }        
    }
    
    
    public void showFrame() {
        if (isVisible()) {
            return;
        }
        
        Window owner = getOwner();
        setLocationRelativeTo(owner);
        setVisible(true);
    }
    
    
    public void hideFrame() {
        dispose();        
    }
    
    
    public void update() {
        logTableModel.fireTableDataChanged();
    }
    
    
    public void clear() {
        logEntries.clear();
    }
    
}
