/*
 * JTopGMap.java
 *
 * Copyright (c) 2005 Bernhard Fischer.
 * This file is part of TopGMap. TopGMap is free software; it is
 * published under the terms of the GNU General Public License.
 */

package de.bfisch.gis.gui;

import de.bfisch.gis.GpsLine;
import de.bfisch.gis.GpsRoute;
import de.bfisch.gis.GpsTrack;
import de.bfisch.gis.GpsWaypoint;
import de.bfisch.gis.TopGMap;
import de.bfisch.gis.TopGMapConfig;
import de.bfisch.io.FileUtils;
import de.bfisch.io.NullOutputStream;
import de.bfisch.swing.NameExtFileFilter;
import de.bfisch.util.Logger;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.Action;
import javax.swing.AbstractAction;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.BorderFactory;
import javax.swing.UIManager;
import javax.swing.WindowConstants;
import javax.swing.filechooser.FileFilter;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;


public class JTopGMap extends JFrame {

    
    class PropertiesAction extends AbstractAction {
        
        public PropertiesAction() {
            super("Einstellungen...", null);
            putValue(SHORT_DESCRIPTION, "ffnet den Einstellungen-Dialog");
            //putValue(MNEMONIC_KEY, mnemonic);
        }
        
        public void actionPerformed(ActionEvent e) {
            showOptionsDialog();
        }
        
    }

    
    class LogAction extends AbstractAction {
        
        public LogAction() {
            super("Protokoll...", null);
            putValue(SHORT_DESCRIPTION, "ffnet das Protokoll-Fenster");
            //putValue(MNEMONIC_KEY, mnemonic);
        }
        
        public void actionPerformed(ActionEvent e) {
            showLogFrame();
        }
        
    }

    
    class ExitAction extends AbstractAction {
        
        public ExitAction() {
            super("Beenden", null);
            putValue(SHORT_DESCRIPTION, "Closes the application");
            //putValue(MNEMONIC_KEY, mnemonic);
        }
        
        public void actionPerformed(ActionEvent e) {
            exit();
        }
        
    }

    
    private static final Logger log = new Logger("JTopGMap");
    
    
    private TopGMapConfig config;
    
    private JFileChooser fileChooser;
    
    private JTextField inputPath;
    private JButton inputButton;
    private FileFilter inputFileFilter;
    
    private JTextField outputPath;
    private JButton outputButton;
    private FileFilter outputFileFilter;
    
    private JButton processButton;
    private JLabel statusLabel;
    
    private GridBagConstraints gbc;

    private Action propertiesAction;
    private Action logAction;
    private Action exitAction;
    
    private JOptionsDialog optionsDialog;
    
    private JLogFrame logFrame;    
    
    
    public JTopGMap() {
        super("TopGMap");
        init();
    }
    
    
    private void applyConfig() {
        int logLevel = config.getLogLevel();
        log.setLevel(logLevel);
        
        String inputFileName = config.getInputPath();
        if (inputFileName != null) {
            inputPath.setText(inputFileName);
        }

        String outputFileName = config.getOutputPath();
        if (outputFileName != null) {
            outputPath.setText(outputFileName);
        }
            
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        Dimension frameSize = getSize();
        if (frameSize.height > screenSize.height) {
            frameSize.height = screenSize.height;
        }
        if (frameSize.width > screenSize.width) {
            frameSize.width = screenSize.width;
        }
        
        int defaultX = (screenSize.width -  frameSize.width ) / 2;
        int x = config.getX(defaultX);
        
        int defaultY = (screenSize.height - frameSize.height) / 2;
        int y = config.getY(defaultY);
        
        setLocation(x, y);
    }
    
    
    private void extractConfig() {
        String inputFileName  = inputPath.getText();
        config.setInputPath(inputFileName);       
        
        String outputFileName = outputPath.getText();
        config.setOutputPath(outputFileName);
        
        int logLevel = log.getLevel();
        config.setLogLevel(logLevel);

        int x = getX();
        config.setX(x);
        
        int y = getY();
        config.setY(y);
    }
    
    
    private void exit() {
        extractConfig();
        
        if (config.getName().equals("JTopGMap")) {
            try {
                config.store();
            } catch (IOException ex) {
                log.warn("Konfiguration kann nicht gespeichert werden",
                	 new Object[] { "Exception", ex });
            }
        }
        this.dispose();
        
        System.exit(0);
    }
    
    
    private void init() {
        PrintStream logStream = new PrintStream(new NullOutputStream());
        log.setLogStream(logStream);
        
        config = TopGMapConfig.getInstance("JTopGMap");
        if (config.getName().equals("JTopGMap")) {
            try {
                config.load();
            } catch (IOException ex) {
                log.warn("Konfiguration kann nicht geladen werden",
                	 new Object[] { "Exception", ex });
            }
        }
        
        propertiesAction = new PropertiesAction();
        logAction = new LogAction();
        exitAction = new ExitAction();

        setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                exit();
            }
        });
        
        JMenuBar menuBar = createMenuBar();
        setJMenuBar(menuBar);
        
        Container contentPane = getContentPane();
        
        Component contentPanel = createMainPane();
        contentPane.add(contentPanel, BorderLayout.CENTER);
        
        applyConfig();        
        
        pack();
        
        createLogFrame();
    }
    
    
    private JMenuBar createMenuBar() {
        JMenuBar menuBar = new JMenuBar();
        
        JMenu fileMenu = new JMenu("Datei");
        menuBar.add(fileMenu);

        JMenuItem propertiesMenuItem = new JMenuItem(propertiesAction);
        fileMenu.add(propertiesMenuItem);

        JMenuItem logMenuItem = new JMenuItem(logAction);
        fileMenu.add(logMenuItem);
        
        fileMenu.addSeparator();

        JMenuItem exitMenuItem = new JMenuItem(exitAction);
        fileMenu.add(exitMenuItem);
        
        JMenu helpMenu = new JMenu("Hilfe");
        menuBar.add(helpMenu);

        JMenuItem aboutMenuItem = new JMenuItem("ber...");
        aboutMenuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                JAboutDialog aboutDialog = new JAboutDialog(JTopGMap.this);
                aboutDialog.showDialog();
            }
        });
        helpMenu.add(aboutMenuItem);
        
        return menuBar;
    }
    
    
    private Component createMainPane() {
        JPanel contentPanel = new JPanel();
        contentPanel.setBorder(BorderFactory.createEmptyBorder(4, 4, 4, 4));
            
        GridBagLayout gridbag = new GridBagLayout();
        contentPanel.setLayout(gridbag);
        
        gbc = new GridBagConstraints();
        gbc.insets = new Insets(6, 2, 3, 2);
        fileChooser = new JFileChooser();
        
        Component openPane = createOpenPane();
        gbc.anchor = GridBagConstraints.CENTER;
        gbc.fill = GridBagConstraints.BOTH;
        gbc.weightx = 1.0;
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gridbag.setConstraints(openPane, gbc);
        contentPanel.add(openPane);
        
        Component savePane = createSavePane();
        gbc.anchor = GridBagConstraints.CENTER;
        gbc.fill = GridBagConstraints.BOTH;
        gbc.weightx = 1.0;
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gridbag.setConstraints(savePane, gbc);
        contentPanel.add(savePane);
        
        Component processPane = createProcessPane();
        gbc.anchor = GridBagConstraints.CENTER;
        gbc.fill = GridBagConstraints.BOTH;
        gbc.weightx = 1.0;
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gridbag.setConstraints(processPane, gbc);
        contentPanel.add(processPane);
        
        return contentPanel;
    }
    
    
    private Component createOpenPane() {
        JPanel openPanel = new JPanel();
        GridBagLayout gridbag = new GridBagLayout();
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.insets = new Insets(3, 2, 3, 2);
        openPanel.setLayout(gridbag);
        openPanel.setBorder(BorderFactory.createTitledBorder("Eingabe (OVL-Datei)"));
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        
        inputPath = new JTextField(20);
        gbc.fill = GridBagConstraints.BOTH;
        gbc.weightx = 1.0;
        gbc.gridwidth = GridBagConstraints.RELATIVE;
        gridbag.setConstraints(inputPath, gbc);
        openPanel.add(inputPath);

        inputButton = new JButton("...");
        gbc.fill = GridBagConstraints.NONE;
        gbc.weightx = 0.0;
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gridbag.setConstraints(inputButton, gbc);
        openPanel.add(inputButton);

        inputFileFilter = new NameExtFileFilter("OVL");

        inputButton.addActionListener(new ActionListener() {
            
            public void actionPerformed(ActionEvent e) {
                String inputFileName = inputPath.getText();
                fileChooser.setSelectedFile(inputFileName != null ? new File(inputFileName) : null);
                fileChooser.resetChoosableFileFilters();
                fileChooser.addChoosableFileFilter(inputFileFilter);
                fileChooser.setFileFilter(inputFileFilter);
                int returnVal = fileChooser.showOpenDialog(JTopGMap.this);

                if (returnVal == JFileChooser.APPROVE_OPTION) {
                    File inputFile = fileChooser.getSelectedFile();
                    FileFilter selectedFileFilter = fileChooser.getFileFilter();
                    if (selectedFileFilter instanceof NameExtFileFilter) {
                        inputFileName = inputFile.getName();
                        NameExtFileFilter selectedNameExtFileFilter = (NameExtFileFilter)selectedFileFilter;
                        String nameExt = selectedNameExtFileFilter.getNameExt().toUpperCase();
                        if (!inputFileName.toUpperCase().endsWith("." + nameExt)) {
                            inputFile = new File(inputFile.getParent(), inputFileName + "." + nameExt);
                        }
                    }
                    inputPath.setText(inputFile.getPath());
                    
                    File currentDir = fileChooser.getCurrentDirectory();
                    config.setCurrentDir(currentDir);
                    
                    String outputFilePath = outputPath.getText();
                    if (outputFilePath.length() == 0)
                        outputFilePath = inputPath.getText();
                    File outputFile = new File(outputFilePath);
                    String inputFileNameCore = FileUtils.coreName(inputFile.getName());
                    outputFilePath = outputFile.getParent() + File.separator + inputFileNameCore + ".gpx";
                    outputPath.setText(outputFilePath);
                }
            }
            
        });
        
        return openPanel;
    }
    
    
    private Component createSavePane() {
        JPanel savePanel = new JPanel();
        GridBagLayout gridbag = new GridBagLayout();
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.insets = new Insets(3, 2, 3, 2);
        savePanel.setLayout(gridbag);
        savePanel.setBorder(BorderFactory.createTitledBorder("Ausgabe (GPX-Datei)"));
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        
        outputPath = new JTextField(20);
        gbc.fill = GridBagConstraints.BOTH;
        gbc.weightx = 1.0;
        gbc.gridwidth = GridBagConstraints.RELATIVE;
        gridbag.setConstraints(outputPath, gbc);
        savePanel.add(outputPath);

        outputButton = new JButton("...");
        gbc.fill = GridBagConstraints.NONE;
        gbc.weightx = 0.0;
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gridbag.setConstraints(outputButton, gbc);
        savePanel.add(outputButton);

        outputFileFilter = new NameExtFileFilter("GPX");
        
        outputButton.addActionListener(new ActionListener() {
            
            public void actionPerformed(ActionEvent e) {
                String outputFileName = outputPath.getText();
                fileChooser.setSelectedFile(outputFileName != null ? new File(outputFileName) : null);
                fileChooser.resetChoosableFileFilters();
                fileChooser.addChoosableFileFilter(outputFileFilter);
                fileChooser.setFileFilter(outputFileFilter);
                int returnVal = fileChooser.showSaveDialog(JTopGMap.this);

                if (returnVal == JFileChooser.APPROVE_OPTION) {
                    File outputFile = fileChooser.getSelectedFile();
                    FileFilter selectedFileFilter = fileChooser.getFileFilter();
                    if (selectedFileFilter instanceof NameExtFileFilter) {
                        outputFileName = outputFile.getName();
                        NameExtFileFilter selectedNameExtFileFilter = (NameExtFileFilter)selectedFileFilter;
                        String nameExt = selectedNameExtFileFilter.getNameExt().toUpperCase();
                        if (!outputFileName.toUpperCase().endsWith("." + nameExt)) {
                            outputFile = new File(outputFile.getParent(), outputFileName + "." + nameExt);
                        }
                    }
                    outputPath.setText(outputFile.getPath());
                    
                    File currentDir = fileChooser.getCurrentDirectory();
                    config.setCurrentDir(currentDir);
                }
            }
            
        });
        
        return savePanel;
    }
    
    
    private Component createProcessPane() {
        JPanel processPanel = new JPanel();
        GridBagLayout gridbag = new GridBagLayout();
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.insets = new Insets(3, 2, 3, 2);
        processPanel.setLayout(gridbag);
        //processPanel.setBorder(BorderFactory.createEmptyBorder());
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        
        processButton = new JButton("Generieren");
        gbc.anchor = GridBagConstraints.CENTER;
        gbc.fill = GridBagConstraints.NONE;
        gbc.weightx = 0.0;
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gridbag.setConstraints(processButton, gbc);
        processPanel.add(processButton);
        
        processButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                process();
            }
        });        
        
        statusLabel = new JLabel(" ");
        //statusLabel.setBorder(BorderFactory.createEmptyBorder(12, 4, 4, 4));
        statusLabel.setBorder(BorderFactory.createCompoundBorder(
                BorderFactory.createTitledBorder(""),
                BorderFactory.createEmptyBorder(3, 2, 3, 2)));
        gbc.fill = GridBagConstraints.BOTH;
        gbc.weightx = 1.0;
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gridbag.setConstraints(statusLabel, gbc);
        processPanel.add(statusLabel);
        
        return processPanel;
    }
    
    
    private void showOptionsDialog() {
	if (optionsDialog == null) {
	    optionsDialog = new JOptionsDialog(this);
    	}
        optionsDialog.showDialog();
    }
    
    
    private void createLogFrame() {
        logFrame = new JLogFrame();
    }
    
    
    private void showLogFrame() {
        logFrame.showFrame();
    }
    
    
    private void clear() {
	log.resetCounter();
	
        logFrame.clear();
        
        statusLabel.setText(" ");
        
        GpsWaypoint.clear();
        GpsLine.clear();
        GpsTrack.clear();
        GpsRoute.clear();
    }
    
    
    private void process() {
        clear();
        
        String inputFileName  = inputPath.getText();
        if (inputFileName.length() == 0) {
            JOptionPane.showMessageDialog(this,
                    "Keine Eingabedatei spezifiziert!",
                    "TopGMap",
                    JOptionPane.ERROR_MESSAGE);
            statusLabel.setText("Generierung abgebrochen.");
            return;
        }
        File inputFile = new File(inputFileName);
        if (!inputFile.isFile() || !inputFile.canRead()) {
            JOptionPane.showMessageDialog(this,
                    "Fehlerhafte Eingabedatei!",
                    "TopGMap",
                    JOptionPane.ERROR_MESSAGE);
            statusLabel.setText("Generierung abgebrochen.");
            return;
        }
        
        String outputFileName = outputPath.getText();
        if (outputFileName.length() == 0) {
            JOptionPane.showMessageDialog(this,
                    "Keine Ausgabedatei spezifiziert!",
                    "TopGMap",
                    JOptionPane.ERROR_MESSAGE);
            statusLabel.setText("Generierung abgebrochen.");
            return;
        }
        String nameExt = FileUtils.nameExt(outputFileName);
        if (nameExt == null || !nameExt.equals("gpx")) {
            outputFileName += ".gpx";
            outputPath.setText(outputFileName);
        }
        File outputFile = new File(outputFileName);
        if (outputFile.isFile() && outputFile.canWrite()) {
            int option = JOptionPane.showConfirmDialog(this,
                    outputFileName + " besteht bereits.\n" +
                    "Mchten Sie diese berschreiben?",
                    "Speichern unter",
                    JOptionPane.YES_NO_OPTION);
            if (option != JOptionPane.YES_OPTION) {
                statusLabel.setText("Generierung abgebrochen.");
                return;
            }
        }
        if (outputFile.exists() && (!outputFile.isFile() || !outputFile.canWrite())) {
            JOptionPane.showMessageDialog(this,
                    "Fehlerhafte Ausgabedatei!",
                    "TopGMap",
                    JOptionPane.ERROR_MESSAGE);
            statusLabel.setText("Generierung abgebrochen.");
            return;
        }
        
        TopGMap topGMap = new TopGMap();
        try {
            log.info("Generierung gestartet.");
            statusLabel.setText("Generierung luft...");
            
            topGMap.process(inputFileName, outputFileName);
            
            topGMap.close();
            
            StringBuffer msg = new StringBuffer("Generierung abgeschlossen");
            String summary = log.getCountSummary();
            if (summary.length() > 0) {
                msg.append(" (");
                msg.append(summary);
                msg.append(")");
            }
            msg.append(".");
            statusLabel.setText(msg.toString());
            log.info("Generierung abgeschlossen.");
            
        } catch (Exception ex) {
            statusLabel.setText("Generierung fehlgeschlagen.");
            log.error("Generierung fehlgeschlagen.", new Object[] { "Exception", ex });
            String message = ex.getMessage();
            //StringWriter bufferWriter = new StringWriter();
            StringWriter buffer = new StringWriter();
            PrintWriter bufferWriter = new PrintWriter(buffer);
            ex.printStackTrace(bufferWriter);
            bufferWriter.close();
            String stackTrace = buffer.toString();
            JOptionPane.showMessageDialog(this,
                    "Generierung fehlgeschlagen: " + (message != null ? message + "!" : ""),
                    "TopGMap",
                    JOptionPane.ERROR_MESSAGE);
        }

        logFrame.update();
    }
    
    
    public static void main(String[] args) {        
        Runnable runner = new Runnable() {         
            public void run() {
                try {
                    UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
                } catch (Exception ex) { /*ignore*/ }
                
                JTopGMap topGMap = new JTopGMap();
                topGMap.setVisible(true);
            }
        };

        EventQueue.invokeLater(runner);
    }
    
}
