/*
 * BasicTableTransferHandler.java
 *
 * Copyright (c) 2005 Bernhard Fischer.
 * This file is part of TopGMap. TopGMap is free software; it is
 * published under the terms of the GNU General Public License.
 */

package de.bfisch.swing;


import java.awt.datatransfer.*;

import javax.swing.JTable;
import javax.swing.TransferHandler;
import javax.swing.JComponent;


public class BasicTableTransferHandler extends TransferHandler {
    
    public static final int FORMAT_VALUE         = 1;
    public static final int FORMAT_ROW_OBJECT    = 2;
    public static final int FORMAT_COLUMN_OBJECT = 3;

    private int formatType;
    
    
    public BasicTableTransferHandler(int formatType) {
        this.formatType = formatType;
    }

    
    /**
     * Create a Transferable to use as the source for a data transfer.
     *
     * @param c  The component holding the data to be transfered.  This
     *  argument is provided to enable sharing of TransferHandlers by
     *  multiple components.
     * @return  The representation of the data to be transfered. 
     *  
     */
    protected Transferable createTransferable(JComponent c) {
    
        if (c instanceof JTable) {
            JTable table = (JTable) c;
            int[] rows;
            int[] cols;
            
            if (!table.getRowSelectionAllowed() && !table.getColumnSelectionAllowed()) {
                return null;
            }
            
            if (!table.getRowSelectionAllowed()) {
                int rowCount = table.getRowCount();

                rows = new int[rowCount];
                for (int counter = 0; counter < rowCount; counter++) {
                    rows[counter] = counter;
                }
            } else {
                rows = table.getSelectedRows();
            }
            
            if (!table.getColumnSelectionAllowed()) {
                int colCount = table.getColumnCount();

                cols = new int[colCount];
                for (int counter = 0; counter < colCount; counter++) {
                    cols[counter] = counter;
                }
            } else {
                cols = table.getSelectedColumns();
            }
            
            if (rows == null || cols == null || rows.length == 0 || cols.length == 0) {
                return null;
            }
            
            StringBuffer plainBuf = new StringBuffer();
            
            switch (formatType) {
            case FORMAT_VALUE:
                for (int idxRow = 0; idxRow < rows.length; idxRow++) {
                    for (int idxCol = 0; idxCol < cols.length; idxCol++) {
                        int row = rows[idxRow];
                        int col = cols[idxCol];
                        Object value = table.getValueAt(row, col);
                        String valueStr = valueToData(table, value, row, col);
                        // Split lines: DOS/Windows/OS/2: "\r\n"; Apple: "\r"; UNIX: "\n"
                        String[] valueLines = valueStr.split("\r\n|\r|\n");
                        if (valueLines.length > 1) {
                            for (int i = 0; i < valueLines.length; i++) {
                                plainBuf.append("\n");
                                plainBuf.append("\t");
                                plainBuf.append(valueLines[i]);
                            }
                        } else {
                            plainBuf.append(valueStr);
                        }
                        plainBuf.append("\t");
                    }
                    // we want a newline at the end of each line and not a tab
                    plainBuf.deleteCharAt(plainBuf.length() - 1).append("\n");
                }
                break;
            case FORMAT_ROW_OBJECT:
                for (int idxRow = 0; idxRow < rows.length; idxRow++) {
                    int row = rows[idxRow];
                    Object object = getObjectAtRow(table, row);
                    String valueStr = objectAtRowToData(table, object, row);
                    plainBuf.append(valueStr);
                    plainBuf.append("\n");
                }
                break;
            case FORMAT_COLUMN_OBJECT:
                for (int idxCol = 0; idxCol < cols.length; idxCol++) {
                    int col = cols[idxCol];
                    Object object = getObjectAtColumn(table, col);
                    String valueStr = objectAtColumnToData(table, object, col);
                    plainBuf.append(valueStr);
                    plainBuf.append("\n");
                }
                break;
            }
            
            // remove the last newline
            plainBuf.deleteCharAt(plainBuf.length() - 1);
            
            Transferable transferable = new StringSelection(plainBuf.toString());
            return transferable;
        }

        return null;
    }


    public int getSourceActions(JComponent c) {
        return COPY;
    }
    
    
    public Object getObjectAtRow(JTable table, int row) {
        assert false : "Method not implemented.";
        return false;
    }
    
    
    public Object getObjectAtColumn(JTable table, int column) {
        assert false : "Method not implemented.";
        return false;
    }
    
    
    public String valueToData(JTable table, Object value, int row, int col) {
        return value != null ? value.toString() : "";
    }
    
    
    public String objectAtRowToData(JTable table, Object obj, int row) {
        return objectToData(table, obj);
    }
    
    
    public String objectAtColumnToData(JTable table, Object obj, int column) {
        return objectToData(table, obj);
    }
    
    
    public String objectToData(JTable table, Object obj) {
        return obj != null ? obj.toString(): "";
    }

}
