/*
 * LogManager.java
 *
 * Copyright (c) 2005 Bernhard Fischer.
 * This file is part of TopGMap. TopGMap is free software; it is
 * published under the terms of the GNU General Public License.
 */

package de.bfisch.util;

import java.io.PrintStream;


/**
 * Logging.
 */
public class LogManager {

    public static int DEBUG = 0;
    public static int INFO  = 1;
    public static int WARN  = 2;
    public static int ERROR = 3;
    public static int FATAL = 4;
    public static int LOG   = 5;

    
    private static class Level {
        
        private int id;
        private String name;
        
        
        public Level(int id, String name) {
            this.id   = id;
            this.name = name;
        }
        
        
        public int getId() {
            return id;
        }
        
        
        public String getName() {
            return name;
        }
        
    }
    
    
    private static final Level[] levels = new Level[] {
        new Level(DEBUG, "DEBUG"),
        new Level(INFO,  "INFO"),
        new Level(WARN,  "WARN"),
        new Level(ERROR, "ERROR"),
        new Level(FATAL, "FATAL"),
        new Level(LOG,   "LOG"),
    };
    
    
    private static LogManager instance;

    private int level = ERROR;
    private LogHandler logHandler = null;
    private PrintStream logStream = System.err;
    private int[] counter = new int[levels.length];
    
    
    public LogManager() {
    }
    
    
    public static LogManager getInstance() {
        if (instance == null) {
            instance = new LogManager();
        }
        return instance;
    }
    
    
    public void setLevel(int level) {
        this.level = level;
    }
    
    
    public int getLevel() {
        return level;
    }
    
    
    public void setLogHandler(LogHandler handler) {
        this.logHandler = handler;
    }
    
    
    public LogHandler getLogHandler() {
        return logHandler;
    }
    
    
    public void setLogStream(PrintStream logStream) {
        this.logStream = logStream != null ? logStream : System.err;
    }
    
    
    public PrintStream getLogStream() {
        return logStream;
    }
    
    
    public void flush() {
        logStream.flush();
    }
    
    
    public static int parseLevel(String levelStr) throws IllegalArgumentException {
        for (int i = 0; i < levels.length; i++) {
            Level level = levels[i];
            if (level.name.equalsIgnoreCase(levelStr)) {
                return level.id;
            }
        }
        throw new IllegalArgumentException("Undefined level string");
    }
    
    
    public String getLevelAsString() {
        return getLevelAsString(level);
    }
    
    
    public static String getLevelAsString(int level) {
        for (int i = 0; i < levels.length; i++) {
            if (levels[i].id == level) {
                return levels[i].name;
            }
        }
        return null;
    }
    
    
    public void log(LogEntry logEntry) {
        if (logHandler != null) {
            logHandler.handle(logEntry);
        }
        
        int level  = logEntry.getLevel();
        String msg = logEntry.getMessage();
        
        counter[level]++;
        
        logStream.print(levels[level].getName());
        logStream.print(": ");
        logStream.println(msg);
        
        Object[] params = logEntry.getParams();
        if (params != null) {
            for (int i = 0; i < params.length; i += 2) {
                Object key   = params[i];
                Object value = params[i + 1];
                logStream.print("  ");
                logStream.print(key.toString());
                logStream.print(": ");
                logStream.println(value.toString());
            }
        }
    }
    
    
    public int getCount(int level) {
        return counter[level];
    }
    
    
    public void resetCounter() {
        for (int level = 0; level < levels.length; level++) {
            counter[level] = 0;
        }
    }
    
}